# Copyright(C) 2017  woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps

import lxml.html

from woob.exceptions import ParseError
from woob.tools.log import getLogger, DEBUG_FILTERS
from woob.tools.misc import NO_DEFAULT as _NO_DEFAULT, NoDefaultType


__all__ = ['FilterError', 'ItemNotFound', 'Filter',]

# Defined for compatibility.
NoDefault = NoDefaultType


class FilterError(ParseError):
    pass


class ItemNotFound(FilterError):
    pass


class _Filter:
    _creation_counter = 0

    def __init__(self, default=_NO_DEFAULT):
        self._key = None
        self._obj = None
        self.default = default
        self._creation_counter = _Filter._creation_counter
        _Filter._creation_counter += 1

    def __or__(self, o):
        self.default = o
        return self

    def __and__(self, o):
        if isinstance(o, type) and issubclass(o, _Filter):
            o = o()
        o.selector = self
        return o

    def __str__(self):
        return self.__class__.__name__

    def __call__(self, item):
        raise NotImplementedError()

    def default_or_raise(self, exception):
        if self.default is not _NO_DEFAULT:
            return self.default
        else:
            raise exception

    def highlight_el(self, el, item=None):
        obj = self._obj or item
        try:
            if not hasattr(obj, 'saved_attrib'):
                return
            if not obj.page.browser.highlight_el:
                return
        except AttributeError:
            return

        if el not in obj.saved_attrib:
            obj.saved_attrib[el] = dict(el.attrib)

        el.attrib['style'] = 'color: white !important; background: red !important;'
        if self._key:
            el.attrib['title'] = 'woob field: %s' % self._key


def debug(*args):
    """
    A decorator function to provide some debug information
    in Filters.
    It prints by default the name of the Filter and the input value.
    """
    def decorator(function):
        logger = getLogger('woob.browser.b2filters')

        def print_debug(self, value):
            result = ''
            outputvalue = value
            if isinstance(value, list):
                from lxml import etree
                outputvalue = ''
                first = True
                for element in value:
                    if first:
                        first = False
                    else:
                        outputvalue += ', '
                    if isinstance(element, etree.ElementBase):
                        outputvalue += "%s" % etree.tostring(element, encoding="unicode")
                    else:
                        outputvalue += "%r" % element
            if self._obj is not None:
                result += "%s" % self._obj._random_id
            if self._key is not None:
                result += ".%s" % self._key
            name = str(self)
            result += " %s(%r" % (name, outputvalue)
            for arg in self.__dict__:
                if arg.startswith('_') or arg == "selector":
                    continue
                if arg == 'default' and getattr(self, arg) == _NO_DEFAULT:
                    continue
                result += ", %s=%r" % (arg, getattr(self, arg))
            result += ')'
            logger.log(DEBUG_FILTERS, result)

        @wraps(function)
        def wrapper(self, value):
            if logger.isEnabledFor(DEBUG_FILTERS):
                print_debug(self, value)

            res = function(self, value)
            return res

        return wrapper
    return decorator


class Filter(_Filter):
    """
    Class used to filter on a HTML element given as call parameter to return
    matching elements.

    Filters can be chained, so the parameter supplied to constructor can be
    either a xpath selector string, or an other filter called before.

    >>> from lxml.html import etree
    >>> f = CleanDecimal(CleanText('//p'), replace_dots=True)
    >>> f(etree.fromstring('<html><body><p>blah: <span>229,90</span></p></body></html>'))
    Decimal('229.90')
    """

    def __init__(self, selector=None, default=_NO_DEFAULT):
        """
        :param default: default value in case the filter fails to find or parse
                        the requested value
        """

        super(Filter, self).__init__(default=default)
        self.selector = selector

    def select(self, selector, item):
        if isinstance(selector, str):
            ret = item.xpath(selector)
        elif isinstance(selector, _Filter):
            selector._key = self._key
            selector._obj = self._obj
            ret = selector(item)
        elif callable(selector):
            ret = selector(item)
        else:
            ret = selector

        if isinstance(ret, lxml.html.HtmlElement):
            self.highlight_el(ret, item)
        elif isinstance(ret, list):
            for el in ret:
                if isinstance(el, lxml.html.HtmlElement):
                    self.highlight_el(el, item)

        return ret

    def __call__(self, item):
        return self.filter(self.select(self.selector, item))

    @debug()
    def filter(self, value):
        """
        This method has to be overridden by children classes.
        """
        raise NotImplementedError()


class _Selector(Filter):
    def filter(self, elements):
        if elements is not None:
            return elements
        else:
            return self.default_or_raise(FilterError('Element %r not found' % self.selector))



# Copyright(C) 2010-2016 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import CapBank

__all__ = [
    'CapBankMatching', 'AccountDiff',
]


class AccountDiff:
    """Difference between 2 accounts lists
    """

    def __init__(self):
        self.matching = []
        """List of new-old account pairs matching together"""

        self.obsolete = []
        """Accounts from the previous state that are not present in the latest state"""

        self.new = []
        """Accounts from the latest state that are not present in the previous state"""

        self.unknown = []


class CapBankMatching(CapBank):
    """
    Capability for matching data between synchronisations.

    This is mostly useful for PFM which have to compare states across time.
    For example, a PFM has to compare accounts freshly returned to accounts
    returned in a previous sync.
    """

    def diff_accounts(self, new_accounts, old_accounts):
        """Compute difference between 2 states of accounts lists.

        This function will remove elements from `new_accounts` and
        `old_accounts` as they are matched and put in the resulting
        `AccountDiff` object.

        Limitations may apply to the fields of `old_accounts` objects, see
        documentation of.

        :param new_accounts: list of freshly fetched, not-matched-yet accounts
        :type new_accounts: iter[:class:`Account`]
        :param old_accounts: list of old, not-matched-yet accounts
        :type old_accounts: iter[:class:`Account`]
        :rtype: iter[:class:`AccountDiff`]
        """

        diff = AccountDiff()
        for new_account in new_accounts:
            old_account = self.match_account(new_account, old_accounts)
            if old_account:
                old_accounts.remove(old_account)
                new_accounts.remove(new_account)
                diff.matching.append((new_account, old_account))
        return diff

    def match_account(self, account, old_accounts):
        """Search an account in `old_accounts` corresponding to `account`.

        `old_accounts` is a list of accounts found in a previous
        synchronisation.
        However, they may not be the exact same objects but only reconstructed
        objects with the same data, although even it could be partial.
        For example, they may have been marshalled, sometimes loosely, thus some
        attributes may be missing (like `_private` attributes) or unset (some
        PFM may choose not to even save all attributes).
        Also, `old_accounts` may not contain all accounts from previous state,
        but only accounts which have not been matched yet.

        :param account: fresh account to search for
        :type account: :class:`Account`
        :param old_accounts: candidates accounts from previous sync
        :type old_accounts: iter[:class:`Account`]
        :return: the corresponding account from `old_accounts`, or `None` if none matches
        :rtype: :class:`Account`
        """

        raise NotImplementedError()



# Copyright(C) 2012-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

# because we don't want to import this file by "import json"
from decimal import Decimal
from datetime import datetime, date, time, timedelta

__all__ = ['json', 'mini_jsonpath']

try:
    # try simplejson first because it is faster
    # However, note that simplejson has very different behaviors from the
    # stdlib json module. In particular, it is handling Decimal in a very
    # peculiar way and is not returning a string for them.
    import simplejson as json
except ImportError:
    # Python 2.6+ has a module similar to simplejson
    import json

from woob.capabilities.base import BaseObject, NotAvailable, NotLoaded


def mini_jsonpath(node, path):
    """
    Evaluates a dot separated path against JSON data. Path can contains
    star wilcards. Always returns a generator.

    Relates to https://goessner.net/articles/JsonPath/ but in a really basic
    and simpler form.

    >>> list(mini_jsonpath({"x": 95, "y": 77, "z": 68}, 'y'))
    [77]
    >>> list(mini_jsonpath({"x": {"y": {"z": "nested"}}}, 'x.y.z'))
    ['nested']
    >>> list(mini_jsonpath('{"data": [{"x": "foo", "y": 13}, {"x": "bar", "y": 42}, {"x": "baz", "y": 128}]}', 'data.*.y'))
    [13, 42, 128]
    """

    def iterkeys(i):
        return range(len(i)) if isinstance(i, list) else i

    def cut(s):
        p = s.split('.', 1) if s else [None]
        return p + [None] if len(p) == 1 else p

    if isinstance(node, str):
        node = json.loads(node)

    queue = [(node, cut(path))]
    while queue:
        node, (name, rest) = queue.pop(0)
        if name is None:
            yield node
            continue
        elif name == '*':
            keys = iterkeys(node)
        elif type(node) not in (dict, list) or name not in node:
            continue
        else:
            keys = [int(name) if type(node) is list else name]
        for k in keys:
            queue.append((node[k], cut(rest)))


class WoobEncoder(json.JSONEncoder):
    """JSON encoder class for woob objects (and Decimal and dates)

    >>> from woob.capabilities.base import BaseObject
    >>> obj = BaseObject(id="1234", backend="my")
    >>> json.dumps(obj, cls=WoobEncoder)
    '{"id": "1234@my", "url": null}'
    """

    def __init__(self, *args, **kwargs):
        # avoid simplejson internal Decimal handling
        if 'use_decimal' in kwargs:
            kwargs['use_decimal'] = False
        super(WoobEncoder, self).__init__(*args, **kwargs)

    def default(self, o):
        if o is NotAvailable:
            return None
        elif o is NotLoaded:
            return None
        elif isinstance(o, BaseObject):
            return o.to_dict()
        elif isinstance(o, Decimal):
            return str(o)
        elif isinstance(o, (datetime, date, time)):
            return o.isoformat()
        elif isinstance(o, timedelta):
            return o.total_seconds()
        return super(WoobEncoder, self).default(o)


WeboobEncoder = WoobEncoder



# Copyright(C) 2010-2013 Romain Bignon, Julien Hébert
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import datetime

from woob.capabilities.base import Currency, empty
from woob.capabilities.travel import CapTravel, RoadmapFilters
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import PrettyFormatter


__all__ = ['AppTravel']


class DeparturesFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'type', 'departure_station', 'arrival_station', 'time')

    def get_title(self, obj):
        s = obj.type
        if hasattr(obj, 'price') and not empty(obj.price):
            s += ' %s %s' % (self.colored('—', 'cyan'), self.colored('%6.2f %s' % (obj.price, Currency.currency2txt(obj.currency)), 'green'))
        if hasattr(obj, 'late') and not empty(obj.late) and obj.late > datetime.time():
            s += ' %s %s' % (self.colored('—', 'cyan'), self.colored('Late: %s' % obj.late, 'red', 'bold'))
        if hasattr(obj, 'information') and not empty(obj.information) and obj.information.strip() != '':
            s += ' %s %s' % (self.colored('—', 'cyan'), self.colored(obj.information, 'red'))
        return s

    def get_description(self, obj):
        if hasattr(obj, 'arrival_time') and not empty(obj.arrival_time):
            s = '(%s)  %s%s\n\t(%s)  %s' % (self.colored(obj.time.strftime('%H:%M') if obj.time else '??:??', 'cyan'),
                                            obj.departure_station,
                                            self.colored(' [Platform: %s]' % obj.platform, 'yellow') if (hasattr(obj, 'platform') and not empty(obj.platform)) else '',
                                            self.colored(obj.arrival_time.strftime('%H:%M'), 'cyan'),
                                            obj.arrival_station)
        else:
            s = '(%s)  %20s -> %s' % (self.colored(obj.time.strftime('%H:%M') if obj.time else '??:??', 'cyan'),
                                      obj.departure_station, obj.arrival_station)

        return s


class StationsFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name')

    def get_title(self, obj):
        return obj.name


class AppTravel(ReplApplication):
    APPNAME = 'travel'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to search for train stations and get departure times."
    SHORT_DESCRIPTION = "search for train stations and departures"
    CAPS = CapTravel
    DEFAULT_FORMATTER = 'table'
    EXTRA_FORMATTERS = {'stations': StationsFormatter,
                        'departures': DeparturesFormatter,
                       }
    COMMANDS_FORMATTERS = {'stations':     'stations',
                           'departures':   'departures',
                          }

    def add_application_options(self, group):
        group.add_option('--departure-time')
        group.add_option('--arrival-time')

    @defaultcount(10)
    def do_stations(self, pattern):
        """
        stations PATTERN

        Search stations.
        """
        for station in self.do('iter_station_search', pattern):
            self.format(station)

    @defaultcount(10)
    def do_departures(self, line):
        """
        departures STATION [ARRIVAL [DATE]]]

        List all departures for a given station.
        The format for the date is "yyyy-mm-dd HH:MM" or "HH:MM".
        """
        station, arrival, date = self.parse_command_args(line, 3, 1)

        station_id, backend_name = self.parse_id(station)
        if arrival:
            arrival_id, backend_name2 = self.parse_id(arrival)
            if backend_name and backend_name2 and backend_name != backend_name2:
                print('Departure and arrival aren\'t on the same backend', file=self.stderr)
                return 1
        else:
            arrival_id = backend_name2 = None

        if backend_name:
            backends = [backend_name]
        elif backend_name2:
            backends = [backend_name2]
        else:
            backends = None

        if date is not None:
            try:
                date = self.parse_datetime(date)
            except ValueError as e:
                print('Invalid datetime value: %s' % e, file=self.stderr)
                print('Please enter a datetime in form "yyyy-mm-dd HH:MM" or "HH:MM".', file=self.stderr)
                return 1

        for departure in self.do('iter_station_departures', station_id, arrival_id, date, backends=backends):
            self.format(departure)

    def do_roadmap(self, line):
        """
        roadmap DEPARTURE ARRIVAL

        Display the roadmap to travel from DEPARTURE to ARRIVAL.

        Command-line parameters:
           --departure-time TIME    requested departure time
           --arrival-time TIME      requested arrival time

        TIME might be in form "yyyy-mm-dd HH:MM" or "HH:MM".

        Example:
            > roadmap Puteaux Aulnay-sous-Bois --arrival-time 22:00
        """
        departure, arrival = self.parse_command_args(line, 2, 2)

        filters = RoadmapFilters()
        try:
            filters.departure_time = self.parse_datetime(self.options.departure_time)
            filters.arrival_time = self.parse_datetime(self.options.arrival_time)
        except ValueError as e:
            print('Invalid datetime value: %s' % e, file=self.stderr)
            print('Please enter a datetime in form "yyyy-mm-dd HH:MM" or "HH:MM".', file=self.stderr)
            return 1

        for route in self.do('iter_roadmap', departure, arrival, filters):
            self.format(route)

    def parse_datetime(self, text):
        if text is None:
            return None

        try:
            date = datetime.datetime.strptime(text, '%Y-%m-%d %H:%M')
        except ValueError:
            try:
                date = datetime.datetime.strptime(text, '%H:%M')
            except ValueError:
                raise ValueError(text)
            date = datetime.datetime.now().replace(hour=date.hour, minute=date.minute)

        return date



# Copyright(C) 2010-2019 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import (
    BaseObject, StringField, DecimalField,
)


class GeoCoordinates(BaseObject):
    latitude = DecimalField('Latitude')
    longitude = DecimalField('Longitude')
    altitude = DecimalField('Altitude')

    def __repr__(self):
        return '<%s %s;%s>' % (type(self).__name__, self.latitude, self.longitude)


class PostalAddress(BaseObject):
    street = StringField('Street address')
    postal_code = StringField('Postal code')
    city = StringField('City')
    region = StringField('Region')
    country = StringField('Country')
    country_code = StringField('ISO 3166-1 alpha-2 country code')

    full_address = StringField('Full address if detailed address is not available')

    def __repr__(self):
        if self.full_address:
            return '<%s full_address=%r>' % (type(self).__name__, self.full_address)
        return '<%s street=%r postal_code=%r region=%r city=%r country=%r country_code=%r>' % (type(self).__name__, self.street, self.postal_code, self.region, self.city, self.country, self.country_code)


def compat_field(field, sub):
    @property
    def f(self):
        return getattr(self, field) and getattr(getattr(self, field), sub)

    @f.setter
    def f(self, value):
        if not getattr(self, field):
            type_ = self._fields[field].types[0]
            setattr(self, field, type_())
        setattr(getattr(self, field), sub, value)

    return f



# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import Capability, BaseObject, NotAvailable, Field, StringField, IntField, Enum
from .date import DateField


__all__ = ['BaseFile', 'CapFile']


class Licenses(Enum):
    OTHER = 'Other license'
    PD = 'Public Domain'
    COPYRIGHT = 'All rights reserved'
    CCBY = 'Creative Commons BY'
    CCBYSA = 'Creative Commons BY-SA'
    CCBYNC = 'Creative Commons BY-NC'
    CCBYND = 'Creative Commons BY-ND'
    CCBYNCSA = 'Creative Commons BY-NC-SA'
    CCBYNCND = 'Creative Commons BY-NC-ND'
    GFDL = 'GNU Free Documentation License'


LICENSES = Licenses


class BaseFile(BaseObject):
    """
    Represent a file.
    """
    title =         StringField('File title')
    ext =           StringField('File extension')
    mime_type =     StringField('MIME Type')
    author =        StringField('File author')
    description =   StringField('File description')
    date =          DateField('File publication date')
    last_update =   DateField('Last update or last modified date', default=NotAvailable)
    size =          IntField('File size in bytes', default=NotAvailable)
    rating =        Field('Rating', int, float, default=NotAvailable)
    rating_max =    Field('Maximum rating', int, float, default=NotAvailable)
    license =       StringField('License name')

    def __str__(self):
        return self.url or ''

    def __repr__(self):
        return '<%s title=%r url=%r>' % (type(self).__name__, self.title, self.url)

    @classmethod
    def id2url(cls, _id):
        """
        Overloaded in child classes provided by backends.
        """
        raise NotImplementedError()

    @property
    def page_url(self):
        """
        Get file page URL
        """
        return self.id2url(self.id)


class SearchSort(Enum):
    RELEVANCE = 0
    RATING = 1
    VIEWS = 2
    DATE = 3


class CapFile(Capability):
    """
    Provide file download
    """
    SEARCH_RELEVANCE = SearchSort.RELEVANCE
    SEARCH_RATING = SearchSort.RATING
    SEARCH_VIEWS = SearchSort.VIEWS
    SEARCH_DATE = SearchSort.DATE

    def search_file(self, pattern, sortby=SEARCH_RELEVANCE):
        """
        :param pattern: pattern to search on
        :type pattern: str
        :param sortby: sort by ... (user SEARCH_* constants)
        :rtype: iter[:class:`BaseFile`]
        """
        raise NotImplementedError()

    def get_file(self, _id):
        """
        Get a file from an ID

        :param _id: the file id. I can be a numeric ID, or a page url
        :type _id: str
        :rtype: :class:`BaseFile` or None if not found.
        """
        raise NotImplementedError()



# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.tools.capabilities.streaminfo import StreamInfo
from .base import Field
from .file import CapFile
from .audio import CapAudio, BaseAudio


__all__ = ['BaseAudioStream', 'CapAudioStream']


class BaseAudioStream(BaseAudio):
    """
    Audio stream object
    """
    current = Field('Information related to current broadcast', StreamInfo)

    def __str__(self):
        return '%s (%s)' % (self.title, self.url)

    def __repr__(self):
        return '%r (%r)' % (self.title, self.url)


class CapAudioStream(CapAudio):
    """
    Audio streams provider
    """

    def search_audiostreams(self, pattern, sortby=CapFile.SEARCH_RELEVANCE):
        """
        Search an audio stream

        :param pattern: pattern to search
        :type pattern: str
        :param sortby: sort by ... (use SEARCH_* constants)
        :rtype: iter[:class:`BaseAudioStream`]
        """
        return self.search_audio(pattern, sortby)

    def get_audiostream(self, _id):
        """
        Get an audio stream

        :param _id: Audiostream ID
        :type id: str
        :rtype: :class:`BaseAudioStream`
        """
        return self.get_audio(_id)



# Copyright(C) 2013 Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from datetime import time, datetime
from dateutil import tz

from woob.tools.date import parse_date
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.capabilities.base import empty
from woob.capabilities.calendar import CapCalendarEvent, Query, CATEGORIES, TICKET, STATUS
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.config.yamlconfig import YamlConfig

__all__ = ['AppCalendar']


class UpcomingSimpleFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'start_date', 'category', 'summary', 'status')

    def format_obj(self, obj, alias):
        result = '%s - %s' % (obj.backend, obj.category)
        if not empty(obj.start_date):
            result += ' - %s' % obj.start_date.strftime('%H:%M')
        result += ' - %s' % obj.summary
        if obj.status == STATUS.CANCELLED:
            result += ' (cancelled)'
        return result


class ICalFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'start_date', 'end_date', 'summary', 'status')

    def start_format(self, **kwargs):
        result = 'BEGIN:VCALENDAR\r\n'
        result += 'VERSION:2.0\r\n'
        result += 'PRODID:-//hacksw/handcal//NONSGML v1.0//EN\r\n'
        self.output(result)

    def format_obj(self, obj, alias):
        result = 'BEGIN:VEVENT\r\n'

        utc_zone = tz.gettz('UTC')

        event_timezone = tz.gettz(obj.timezone)
        start_date = obj.start_date if not empty(obj.start_date) else datetime.now()
        if isinstance(start_date, datetime):
            start_date = start_date.replace(tzinfo=event_timezone)
            utc_start_date = start_date.astimezone(utc_zone)
            result += 'DTSTART:%s\r\n' % utc_start_date.strftime("%Y%m%dT%H%M%SZ")
        else:
            result += 'DTSTART:%s\r\n' % start_date.strftime("%Y%m%d")

        end_date = obj.end_date if not empty(obj.end_date) else datetime.combine(start_date, time.max)
        if isinstance(end_date, datetime):
            end_date = end_date.replace(tzinfo=event_timezone)
            utc_end_date = end_date.astimezone(utc_zone)
            result += 'DTEND:%s\r\n' % utc_end_date.strftime("%Y%m%dT%H%M%SZ")
        else:
            result += 'DTEND:%s\r\n' % end_date.strftime("%Y%m%d")

        result += 'SUMMARY:%s\r\n' % obj.summary
        result += 'UID:%s\r\n' % obj.id
        result += 'STATUS:%s\r\n' % obj.status

        location = ''
        if hasattr(obj, 'location') and not empty(obj.location):
            location += obj.location + ' '

        if hasattr(obj, 'city') and not empty(obj.city):
            location += obj.city + ' '

        if not empty(location):
            result += 'LOCATION:%s\r\n' % location

        if hasattr(obj, 'categories') and not empty(obj.categories):
            result += 'CATEGORIES:%s\r\n' % obj.categories

        if hasattr(obj, 'description') and not empty(obj.description):
            result += 'DESCRIPTION:%s\r\n' % obj.description.strip(' \t\n\r')\
                                                             .replace('\r', '')\
                                                             .replace('\n', r'\n')\
                                                             .replace(',', r'\,')

        if hasattr(obj, 'transp') and not empty(obj.transp):
            result += 'TRANSP:%s\r\n' % obj.transp

        if hasattr(obj, 'sequence') and not empty(obj.sequence):
            result += 'SEQUENCE:%s\r\n' % obj.sequence

        if hasattr(obj, 'url') and not empty(obj.url):
            result += 'URL:%s\r\n' % obj.url

        result += 'END:VEVENT\r\n'
        return result

    def flush(self, **kwargs):
        self.output('END:VCALENDAR')


class UpcomingListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'start_date', 'end_date', 'summary', 'category', 'status')

    def get_title(self, obj):
        return ' %s - %s ' % (obj.category, obj.summary)

    def get_description(self, obj):
        result = ''
        if not empty(obj.start_date):
            result += '\tDate: %s\n' % obj.start_date.strftime('%A %d %B %Y')
            result += '\tHour: %s' % obj.start_date.strftime('%H:%M')
            if not empty(obj.end_date):
                result += ' - %s' % obj.end_date.strftime('%H:%M')
                days_diff = (obj.end_date - obj.start_date).days
                if days_diff >= 1:
                    result += ' (%i day(s) later)' % (days_diff)
            result += '\n'
        if obj.status == STATUS.CANCELLED:
            result += '\tStatus: Cancelled\n'
        return result.strip('\n\t')


class UpcomingFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'start_date', 'end_date', 'summary', 'category', 'status')

    def format_obj(self, obj, alias):
        result = '%s%s - %s%s\n' % (self.BOLD, obj.category, obj.summary, self.NC)

        if not empty(obj.start_date):
            if not empty(obj.end_date):
                days_diff = (obj.end_date - obj.start_date).days
                if days_diff >= 1:
                    result += 'From: %s to %s ' % (obj.start_date.strftime('%A %d %B %Y'),
                                                    obj.end_date.strftime('%A %d %B %Y'))
                else:
                    result += 'Date: %s\n' % obj.start_date.strftime('%A %d %B %Y')
                    result += 'Hour: %s' % obj.start_date.strftime('%H:%M')
                    result += ' - %s' % obj.end_date.strftime('%H:%M')
            else:
                result += 'Date: %s\n' % obj.start_date.strftime('%A %d %B %Y')
                result += 'Hour: %s' % obj.start_date.strftime('%H:%M')

            result += '\n'

        if hasattr(obj, 'location') and not empty(obj.location):
            result += 'Location: %s\n' % obj.location

        if hasattr(obj, 'city') and not empty(obj.city):
            result += 'City: %s\n' % obj.city

        if hasattr(obj, 'event_planner') and not empty(obj.event_planner):
            result += 'Event planner: %s\n' % obj.event_planner

        if hasattr(obj, 'booked_entries') and not empty(obj.booked_entries) and \
           hasattr(obj, 'max_entries') and not empty(obj.max_entries):
            result += 'Entry: %s/%s \n' % (obj.booked_entries, obj.max_entries)
        elif hasattr(obj, 'booked_entries') and not empty(obj.booked_entries):
            result += 'Entry: %s \n' % (obj.booked_entries)
        elif hasattr(obj, 'max_entries') and not empty(obj.max_entries):
            result += 'Max entries: %s \n' % (obj.max_entries)

        if hasattr(obj, 'description') and not empty(obj.description):
            result += 'Description:\n %s\n\n' % obj.description

        if hasattr(obj, 'price') and not empty(obj.price):
            result += 'Price: %.2f\n' % obj.price

        if hasattr(obj, 'ticket') and not empty(obj.ticket):
            result += 'Ticket: %s\n' % obj.ticket

        if hasattr(obj, 'url') and not empty(obj.url):
            result += 'URL: %s\n' % obj.url

        if hasattr(obj, 'status') and not empty(obj.status):
            result += 'Status: %s\n' % obj.status

        return result


class AppCalendar(ReplApplication):
    APPNAME = 'calendar'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Bezleputh'
    DESCRIPTION = "Console application to see upcoming events."
    SHORT_DESCRIPTION = "see upcoming events"
    CAPS = CapCalendarEvent
    EXTRA_FORMATTERS = {'upcoming_list': UpcomingListFormatter,
                        'upcoming': UpcomingFormatter,
                        'simple_upcoming': UpcomingSimpleFormatter,
                        'ical_formatter': ICalFormatter,
                        }
    COMMANDS_FORMATTERS = {'list': 'upcoming_list',
                           'search': 'upcoming_list',
                           'load': 'upcoming_list',
                           'ls': 'upcoming_list',
                           'info': 'upcoming',
                           'export': 'ical_formatter'
                           }

    def main(self, argv):
        self.load_config(klass=YamlConfig)
        return super().main(argv)

    def select_values(self, values_from, values_to, query_str):
        r = 'notempty'
        while r != '':
            for i, value in enumerate(values_from, 1):
                print('  %s%2d)%s [%s] %s' % (self.BOLD,
                                              i,
                                              self.NC,
                                              'x' if value in values_to else ' ',
                                              value))
            r = self.ask(query_str, regexp=r'(\d+|)', default='')

            if not r.isdigit():
                continue
            r = int(r)
            if r <= 0 or r > len(values_from):
                continue
            value = list(values_from)[r - 1]
            if value in values_to:
                values_to.remove(value)
            else:
                values_to.append(value)

    @defaultcount(10)
    def do_search(self, line):
        """
        search

        search for an event. Parameters interactively asked
        """
        query = Query()
        self.select_values(CATEGORIES, query.categories, '  Select category (or empty to stop)')
        self.select_values(TICKET, query.ticket, '  Select tickets status (or empty to stop)')

        if query.categories and len(query.categories) > 0 and query.ticket and len(query.ticket) > 0:
            query.city = self.ask('Enter a city', default='')
            query.summary = self.ask('Enter a title', default='')

            start_date = self.ask_date('Enter a start date', default='today')
            end_date = self.ask_date('Enter a end date', default='')

            if end_date:
                if end_date == start_date:
                    end_date = datetime.combine(start_date, time.max)
                else:
                    end_date = datetime.combine(end_date, time.max)

            query.start_date = datetime.combine(start_date, time.min)
            query.end_date = end_date

            save_query = self.ask('Save query (y/n)', default='n')
            if save_query.upper() == 'Y':
                name = ''
                while not name:
                    name = self.ask('Query name')

                self.config.set('queries', name, query)
                self.config.save()
            self.complete_search(query)

    def complete_search(self, query):
        self.change_path(['events'])
        self.start_format()
        for event in self.do('search_events', query):
            if event:
                self.cached_format(event)

    def ask_date(self, txt, default=''):
        r = self.ask(txt, default=default)
        return parse_date(r)

    @defaultcount(10)
    def do_load(self, query_name):
        """
        load [query name]
        without query name : list loadable queries
        with query name laod query
        """
        queries = self.config.get('queries')
        if not queries:
            print('There is no saved queries', file=self.stderr)
            return 2

        if not query_name:
            for name in queries.keys():
                print('  %s* %s %s' % (self.BOLD,
                                       self.NC,
                                       name))
            query_name = self.ask('Which one')

        if query_name in queries:
            self.complete_search(queries.get(query_name))
        else:
            print('Unknown query', file=self.stderr)
            return 2

    @defaultcount(10)
    def do_list(self, line):
        """
        list [PATTERN]
        List upcoming events, pattern can be an english or french week day, 'today' or a date (dd/mm/yy[yy])
        """

        self.change_path(['events'])
        if line:
            _date = parse_date(line)
            if not _date:
                print('Invalid argument: %s' % self.get_command_help('list'), file=self.stderr)
                return 2

            date_from = datetime.combine(_date, time.min)
            date_to = datetime.combine(_date, time.max)
        else:
            date_from = datetime.now()
            date_to = None

        for event in self.do('list_events', date_from, date_to):
            self.cached_format(event)

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, _id):
        """
        info ID

        Get information about an event.
        """

        if not _id:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        event = self.get_object(_id, 'get_event')

        if not event:
            print('Upcoming event not found: %s' % _id, file=self.stderr)
            return 3

        self.start_format()
        self.format(event)

    def do_export(self, line):
        """
        export FILENAME [ID1 ID2 ID3 ...]

        ID is the identifier of the event. If no ID every events are exported

        FILENAME is where to write the file. If FILENAME is '-', the file is written to stdout.

        Export event in ICALENDAR format
        """
        if not line:
            print('This command takes at leat one argument: %s' % self.get_command_help('export'), file=self.stderr)
            return 2

        _file, args = self.parse_command_args(line, 2, req_n=1)

        l = self.retrieve_events(args)
        if l == 3:
            return 3

        if not _file == "-":
            dest = self.check_file_ext(_file)
            self.formatter.outfile = dest

        self.formatter.start_format()
        for item in l:
            self.format(item)

    def retrieve_events(self, args):
        l = []

        if not args:
            _ids = []
            for event in self.do('list_events', datetime.now(), None):
                _ids.append(event.id)
        else:
            _ids = args.strip().split(' ')

        for _id in _ids:
            event = self.get_object(_id, 'get_event')

            if not event:
                print('Upcoming event not found: %s' % _id, file=self.stderr)
                return 3

            l.append(event)

        return l

    def check_file_ext(self, _file):
        splitted_file = _file.split('.')
        if splitted_file[-1] != 'ics':
            return "%s.ics" % _file
        else:
            return _file

    def do_attends(self, line):
        """
        attends ID1 [ID2 ID3 ...]

        Register as participant of an event.
        ID is the identifier of the event.
        """
        if not line:
            print('This command takes at leat one argument: %s' % self.get_command_help('attends'), file=self.stderr)
            return 2

        args = self.parse_command_args(line, 1, req_n=1)

        l = self.retrieve_events(args[0])
        for event in l:
            # we wait till the work be done, else the errors are not handled
            self.do('attends_event', event, True).wait()

    def do_unattends(self, line):
        """
        unattends ID1 [ID2 ID3 ...]

        Unregister you participation for an event.
        ID is the identifier of the event.
        """

        if not line:
            print('This command takes at leat one argument: %s' % self.get_command_help('unattends'), file=self.stderr)
            return 2

        args = self.parse_command_args(line, 1, req_n=1)

        l = self.retrieve_events(args[0])
        for event in l:
            self.do('attends_event', event, False)



# Copyright(C) 2013-2021 Julien Hebert
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.tools.json import json, WoobEncoder

from .iformatter import IFormatter

__all__ = ['JsonFormatter', 'JsonLineFormatter']


class JsonFormatter(IFormatter):
    """
    Formats the whole list as a single JSON list object.
    """

    def __init__(self):
        super(JsonFormatter, self).__init__()
        self.queue = []

    def flush(self):
        self.output(json.dumps(self.queue, cls=WoobEncoder))

    def format_dict(self, item):
        self.queue.append(item)

    def format_collection(self, collection, only):
        self.queue.append(collection.to_dict())


class JsonLineFormatter(IFormatter):
    """
    Formats the list as received, with a JSON object per line.
    The advantage is that it can be streamed.
    """

    def format_dict(self, item):
        self.output(json.dumps(item, cls=WoobEncoder))



# Copyright(C) 2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import warnings
import importlib

from datetime import datetime
from functools import wraps
from typing import List, Any, Tuple

from woob.tools.value import Value


__all__ = [
    'BrowserIncorrectPassword',
    'BrowserForbidden',
    'BrowserUserBanned',
    'BrowserUnavailable',
    'ScrapingBlocked',
    'BrowserInteraction',
    'BrowserQuestion',
    'OTPQuestion',
    'OTPSentType',
    'SentOTPQuestion',
    'OfflineOTPQuestion',
    'DecoupledMedium',
    'DecoupledValidation',
    'AppValidation',
    'AppValidationError',
    'AppValidationCancelled',
    'AppValidationExpired',
    'BrowserRedirect',
    'CaptchaQuestion',
    'WrongCaptchaResponse',
    'BrowserHTTPNotFound',
    'BrowserHTTPError',
    'BrowserHTTPSDowngrade',
    'BrowserSSLError',
    'ParseError',
    'FormFieldConversionWarning',
    'ModuleInstallError',
    'ModuleLoadError',
    'ActionType',
    'ActionNeeded',
    'AuthMethodNotImplemented',
    'BrowserPasswordExpired',
    'NeedInteractive',
    'NeedInteractiveForRedirect',
    'NeedInteractiveFor2FA',
    'NotImplementedWebsite',
]


class BrowserIncorrectPassword(Exception):
    """The site signals to us our credentials are invalid.

    :type message: str
    :param message: compatibility message for the user (mostly when bad_fields is not given)
    :type bad_fields: list[str]
    :param bad_fields: list of config field names which are incorrect, if it is known
    """

    def __init__(
        self,
        message: str = "",
        bad_fields: List[str] | None = None
    ):
        super().__init__(*filter(None, [message]))
        self.bad_fields = bad_fields


class BrowserForbidden(Exception):
    """The site signals to us that access to a resource is forbidden."""


class BrowserUserBanned(BrowserIncorrectPassword):
    """The site signals to us the user we are logging in as is banned.

    :type message: str
    :param message: compatibility message for the user (mostly when bad_fields is not given)
    :type bad_fields: list[str]
    :param bad_fields: list of config field names which are incorrect, if it is known
    """


class BrowserUnavailable(Exception):
    """The site is either momentarily unavailable, or in maintenance."""


class ScrapingBlocked(BrowserUnavailable):
    """The site has detected scraping, and signals that it has blocked it."""


class BrowserInteraction(Exception):
    """Base class for most browser interactions."""


class BrowserQuestion(BrowserInteraction):
    """The site requires values to be provided by the end user.

    :param fields: The Value objects to be provided by the end user.
    """
    def __init__(self, *fields):
        super().__init__()
        self.fields = fields

    def __str__(self):
        return ", ".join("{}: {}".format(
            field.id or field.label, field.description) for field in self.fields
        )


class OTPQuestion(BrowserQuestion):
    """The site requires transient values to be provided by the end user.

    :param fields: The Value objects to be provided by the end user.
    """


class OTPSentType:
    UNKNOWN = 'unknown'
    SMS = 'sms'
    PHONE_CALL = 'phone_call'
    MOBILE_APP = 'mobile_app'
    EMAIL = 'email'
    DEVICE = 'device'


class SentOTPQuestion(OTPQuestion):
    """A one-time password sent to one of the end user's device is required.

    For example, the site has sent an SMS including a one-time password to a
    phone number associated with the end user, and we need to ask the end
    user to provide this code to us to send it back to the site.

    :type field_name: str
    :param field_name: name of the config field in which the OTP shall
                       be given to the module
    :type medium_type: OTPSentType
    :param medium_type: if known, where the OTP was sent
    :type medium_label: str
    :param medium_label: if known, label of where the OTP was sent,
                         e.g. the phone number in case of an SMS
    :type message: str
    :param message: compatibility message (used as the Value label)
    :type expires_at: datetime.datetime
    :param expires_at: date when the OTP expires and when replying is too late
    """

    def __init__(
        self,
        field_name: str,
        medium_type: str = OTPSentType.UNKNOWN,
        medium_label: str | None = None,
        message: str = "",
        expires_at: datetime | None = None,
    ):
        super().__init__(Value(field_name, label=message))
        self.message = message
        self.medium_type = medium_type
        self.medium_label = medium_label
        self.expires_at = expires_at


class OfflineOTPQuestion(OTPQuestion):
    """A one-time password generated by the end user is required.

    For example, the site requires a one-time password generated by an
    RFC 6238 compliant application, such as Google Authenticator, to be
    provided, so this exception gets raised to obtain one we can send back
    to the site.

    :type field_name: str
    :param field_name: name of the config field in which the OTP shall
                       be given to the module
    :type input: str
    :param input: if relevant, input data for computing the OTP
    :type message: str
    :param message: compatibility message (used as the Value label)
    :type medium_label: str
    :param medium_label: if known, label of the device to use for generating
                         or reading the OTP, e.g. the card index for paper OTP
    :type expires_at: datetime.datetime
    :param expires_at: date when the OTP expires and when replying is too late
    """

    def __init__(
        self,
        field_name: str,
        input: str | None = None,
        medium_label: str | None = None,
        message: str = "",
        expires_at: datetime | None = None
    ):
        super().__init__(Value(field_name, label=message))
        self.input = input
        self.medium_label = medium_label
        self.expires_at = expires_at


class DecoupledMedium:
    UNKNOWN = 'unknown'
    SMS = 'sms'
    MOBILE_APP = 'mobile_app'
    EMAIL = 'email'


class DecoupledValidation(BrowserInteraction):
    """A validation of the current action is requested on a separate channel.

    For example, the site requires the user to click on a link sent in an
    e-mail to pursue logging in on the current session.

    :type medium_type: DecoupledMedium
    :param medium_type: if known, where the decoupled validation was sent
    :type medium_label: str
    :param medium_label: if known, label of where the decoupled validation was
                         sent, e.g. the phone number in case of an app
    :type expires_at: datetime.datetime
    :param expires_at: date when the OTP expires and when replying is too late
    """
    def __init__(
        self,
        message: str = '',
        resource: Any = None,
        medium_type: str = DecoupledMedium.UNKNOWN,
        medium_label: str | None = None,
        expires_at: datetime | None = None,
        *values: Any
    ):
        """
        :param message: message to display to user
        :type message: str
        :param medium_type: if known, where the decoupled validation was sent
        :type medium_type: DecoupledMedium
        :param medium_label: if known, label of where the decoupled validation was sent,
                             e.g. the phone number in case of an app
        :type medium_label: str
        :param expires_at: date when the OTP expires and when replying is too late
        :type expires_at: datetime.datetime
        """
        if values:
            warnings.warn(
                'Variable arguments will be removed in woob 4.',
                DeprecationWarning,
                stacklevel=2
            )
        if resource:
            warnings.warn(
                'The "resource" argument will be removed in woob 4. '
                'Maybe you should inherit this exception class in a '
                'capability to add specific metadata?',
                DeprecationWarning,
                stacklevel=2
            )

        super().__init__(*values)
        self.medium_type = medium_type
        self.medium_label = medium_label
        self.message = message
        self.resource = resource
        self.expires_at = expires_at

    def __str__(self):
        return self.message


class AppValidation(DecoupledValidation):
    """A validation of the current action is requested in a mobile application.

    For example, the site requires the user to open the mobile application
    corresponding to the site, enter a specific password to the application,
    and click on "Validate" for the current operation to be validated.

    :type medium_label: str
    :param medium_label: if known, label of where the decoupled validation was
                         sent, e.g. the phone number in case of an app
    :type expires_at: datetime.datetime
    :param expires_at: date when the OTP expires and when replying is too late
    """

    def __init__(self, *args, **kwargs):
        kwargs['medium_type'] = DecoupledMedium.MOBILE_APP
        super().__init__(*args, **kwargs)


class AppValidationError(Exception):
    """The mobile application validation has failed for a generic reason."""


class AppValidationCancelled(AppValidationError):
    """The mobile application validation has been cancelled.

    This usually happens when the end user has selected "Deny" or "Cancel"
    instead of "Validate" in their mobile application.
    """


class AppValidationExpired(AppValidationError):
    """The mobile application validation has expired.

    This usually happens when the end user hasn't selected any option,
    logged into their mobile application in time to select any option,
    or if they have given up consenting to the operation and the application
    doesn't have a "Deny" option.
    """


class BrowserRedirect(BrowserInteraction):
    """The site requires the end user to do a webauth redirect flow.

    The exception specifies the URL to redirect the user to. At the end
    of the process, i.e. when the end user has been redirected to the
    callback URI with parameters, the 'auth_uri' configuration value should
    be set to the callback URI with parameters for processing by the module.

    :param url: The URL to redirect the end user to.
    :type url: str
    """
    def __init__(
        self,
        url: str,
        resource: Any = None
    ):
        self.url = url

        if resource:
            warnings.warn(
                'The "resource" argument will be removed in woob 4. '
                'Maybe you should inherit this exception class in a '
                'capability to add specific metadata?',
                DeprecationWarning,
                stacklevel=2
            )

        self.resource = resource

    def __str__(self):
        return 'Redirecting to %s' % self.url


class CaptchaQuestion(Exception):
    """The site requires solving a CAPTCHA (base class).

    The response to the captcha should be set, as text, to the
    'captcha_response' configuration value.

    :param type: The type of captcha, as a string.
    """

    # could be improved to pass the name of the backendconfig key
    def __init__(
        self,
        type: str | None = None,
        **kwargs
    ):
        super().__init__('The site requires solving a captcha')
        self.type = type
        for key, value in kwargs.items():
            setattr(self, key, value)


class WrongCaptchaResponse(Exception):
    """The site signals to us that our captcha response is incorrect."""

    def __init__(
        self,
        message: str | None = None
    ):
        super().__init__(message or "Captcha response is wrong")


class BrowserHTTPNotFound(Exception):
    pass


class BrowserHTTPError(Exception):
    pass


class BrowserHTTPSDowngrade(Exception):
    pass


class BrowserSSLError(BrowserUnavailable):
    pass


class ParseError(Exception):
    pass


class FormFieldConversionWarning(UserWarning):
    """
    A value has been set to a form's field and has been implicitly converted.
    """


class ModuleInstallError(Exception):
    pass


class ModuleLoadError(Exception):
    def __init__(
        self,
        module_name: str,
        msg: str
    ):
        super().__init__(msg)
        self.module = module_name


class ActionType:
    # TODO use enum class
    ACKNOWLEDGE = 1
    """Must acknowledge new Terms of Service or some important message"""

    FILL_KYC = 2
    """User information must be filled on website"""

    ENABLE_MFA = 3
    """MFA must be enabled on website"""

    PERFORM_MFA = 4
    """Must perform MFA on website directly to unlock scraping

    It is different from `DecoupledValidation`.
    """

    PAYMENT = 5
    """Must pay site for the feature or pay again for the subscription which has ended"""

    CONTACT = 6
    """Must contact site support or a customer relation person for another problem

    The problem should ideally be described in `ActionNeeded.message`.
    """


class ActionNeeded(Exception):
    """An action must be performed directly, often on website.

    :param message: message from the site
    :type message: str
    :param locale: ISO4646 language tag of `message` (e.g. "en-US")
    :type locale: str
    :param action_type: type of action to perform
    :param url: URL of the page to go to resolve the action needed
    :type url: str
    :param page: user hint for when no URL can be given and the place where to perform the action is not obvious
    :type page: str
    """

    def __init__(
        self,
        message: str | None = None,
        *,
        locale: str | None = None,
        action_type: int | None = None,
        url: str | None = None,
        page: Any = None,
    ):
        args: Tuple[str, ...] = ()
        if message:
            args = (message,)

        super().__init__(*args)
        self.locale = locale
        self.action_type = action_type
        self.page = page
        self.url = url

        if page:
            warnings.warn(
                'ActionNeeded.page is deprecated and will be removed',
                DeprecationWarning,
                stacklevel=2
            )


class AuthMethodNotImplemented(ActionNeeded):
    """Website requires a kind of authentication that is not implemented."""


class BrowserPasswordExpired(ActionNeeded):
    """Credentials are expired, user has to go on website to update them."""


class NeedInteractive(Exception):
    """Require an interactive call by user.

    This may be raised when a method is called by a background job, without a user.
    """


class NeedInteractiveForRedirect(NeedInteractive):
    """Require an interactive call by user to perform a redirect."""


class NeedInteractiveFor2FA(NeedInteractive):
    """Require an interactive call by user to perform a 2FA."""


def implemented_websites(*cfg):
    """
    Decorator to raise NotImplementedWebsite for concerned website
    Will raise the exception for website not in arguments: ex ('ent', 'pro')
    """
    warnings.warn(
        'Do not use this decorator.',
        DeprecationWarning,
        stacklevel=2
    )

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.config['website'].get() in cfg:
                raise NotImplementedWebsite('This website is not yet implemented')

            return func(self, *args, **kwargs)
        return wrapper
    return decorator


class NotImplementedWebsite(NotImplementedError):
    """
    Exception for modules when a website is not yet available.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'Do not use this exception.',
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)


__deprecated__ = {
    'ImageCaptchaQuestion': 'woob.capabilities.captcha.ImageCaptchaQuestion',
    'RecaptchaV2Question': 'woob.capabilities.captcha.RecaptchaV2Question',
    'RecaptchaQuestion': 'woob.capabilities.captcha.RecaptchaQuestion',
    'GeetestV4Question': 'woob.capabilities.captcha.GeetestV4Question',
    'RecaptchaV3Question': 'woob.capabilities.captcha.RecaptchaV3Question',
    'FuncaptchaQuestion': 'woob.capabilities.captcha.FuncaptchaQuestion',
    'HcaptchaQuestion': 'woob.capabilities.captcha.HcaptchaQuestion',
    'TurnstileQuestion': 'woob.capabilities.captcha.TurnstileQuestion',
    'NoAccountsException': 'woob.capabilities.bank.NoAccountsException',
}


def __getattr__(name: str) -> Exception:
    if name in __deprecated__:
        new_path = __deprecated__[name]
        warnings.warn(
            f"'{name}' is deprecated. Use '{new_path}' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        module_name = '.'.join(new_path.split('.')[:-1])
        class_name = new_path.split('.')[-1]
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as exc:
            raise AttributeError(f"{name} is deprecated, but unable to import {new_path}") from exc

        try:
            return getattr(module, class_name)
        except AttributeError as exc:
            raise AttributeError(f"{name} is deprecated, but unable to import {new_path}") from exc

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__() -> List[str]:
    return sorted(list(__all__) + list(__deprecated__.keys()))



# Copyright(C) 2013  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.base import empty
from woob.capabilities.parcel import CapParcel, Parcel, ParcelNotFound
from woob.core import CallErrors
from woob.tools.application.repl import ReplApplication
from woob.tools.application.formatters.iformatter import IFormatter


__all__ = ['AppParcel']


STATUS = {Parcel.STATUS_PLANNED:    ('PLANNED', 'red'),
          Parcel.STATUS_IN_TRANSIT: ('IN TRANSIT', 'yellow'),
          Parcel.STATUS_ARRIVED:    ('ARRIVED', 'green'),
          Parcel.STATUS_UNKNOWN:    ('', 'white'),
         }


def get_backend_name(backend):
    return backend.name


class HistoryFormatter(IFormatter):
    MANDATORY_FIELDS = ()

    def format_obj(self, obj, alias):
        if isinstance(obj, Parcel):
            result =  'Parcel %s (%s)\n' % (self.colored(obj.id, 'red', 'bold'),
                                              self.colored(obj.backend, 'blue', 'bold'))
            result += '%sArrival:%s %s\n' % (self.BOLD, self.NC, obj.arrival)
            status, status_color = STATUS[obj.status]
            result += '%sStatus:%s  %s\n' % (self.BOLD, self.NC, self.colored(status, status_color))
            result += '%sInfo:%s  %s\n\n' % (self.BOLD, self.NC, obj.info)
            result += ' Date                  Location          Activity                                          \n'
            result += '---------------------+-----------------+---------------------------------------------------'
            return result

        return ' %s   %s %s' % (self.colored('%-19s' % obj.date, 'blue'),
                                self.colored('%-17s' % (obj.location or ''), 'magenta'),
                                self.colored(obj.activity or '', 'yellow'))


class StatusFormatter(IFormatter):
    MANDATORY_FIELDS = ('id',)

    def format_obj(self, obj, alias):
        if alias is not None:
            id = '%s (%s)' % (self.colored('%3s' % ('#' + alias), 'red', 'bold'),
                              self.colored(obj.backend, 'blue', 'bold'))
            clean = '#%s (%s)' % (alias, obj.backend)
            if len(clean) < 15:
                id += (' ' * (15 - len(clean)))
        else:
            id = self.colored('%30s' % obj.fullid, 'red', 'bold')

        status, status_color = STATUS[obj.status]
        arrival = obj.arrival.strftime('%Y-%m-%d') if not empty(obj.arrival) else ''
        result = '%s %s %s %s  %s' % (id, self.colored('—', 'cyan'),
                                       self.colored('%-10s' % status, status_color),
                                       self.colored('%-10s' % arrival, 'blue'),
                                       self.colored('%-20s' % obj.info, 'yellow'))

        return result


class AppParcel(ReplApplication):
    APPNAME = 'parcel'
    OLD_APPNAME = 'parceloob'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Romain Bignon'
    CAPS = CapParcel
    DESCRIPTION = "Console application to track your parcels."
    SHORT_DESCRIPTION = "manage your parcels"
    EXTRA_FORMATTERS = {'status':   StatusFormatter,
                        'history':  HistoryFormatter,
                       }
    DEFAULT_FORMATTER = 'table'
    COMMANDS_FORMATTERS = {'status':      'status',
                           'info':        'history',
                          }
    STORAGE = {'tracking': []}

    def do_track(self, line):
        """
        track ID

        Track a parcel.
        """
        parcel = self.get_object(line, 'get_parcel_tracking')
        if not parcel:
            print('Error: the parcel "%s" is not found' % line, file=self.stderr)
            return 2

        parcels = set(self.storage.get('tracking', default=[]))
        parcels.add(parcel.fullid)
        self.storage.set('tracking', list(parcels))
        self.storage.save()

        print('Parcel "%s" has been tracked.' % parcel.fullid)

    def do_untrack(self, line):
        """
        untrack ID

        Stop tracking a parcel.
        """
        removed = False
        # Always try to first remove the parcel, the untrack should always success
        parcels = set(self.storage.get('tracking', default=[]))
        try:
            parcels.remove(line)
            removed = True
        except KeyError:
            pass

        if not removed:
            try:
                parcel = self.get_object(line, 'get_parcel_tracking')
            except ParcelNotFound:
                parcel = False

            if not parcel:
                print('Error: the parcel "%s" is not found. Did you provide the full id@backend parameter?' % line, file=self.stderr)
                return 2

            try:
                parcels.remove(parcel.fullid)
            except KeyError:
                print("Error: parcel \"%s\" wasn't tracked" % parcel.fullid, file=self.stderr)
                return 2

        self.storage.set('tracking', list(parcels))
        self.storage.save()

        if removed:
            print("Parcel \"%s\" isn't tracked anymore." % line)
        else:
            print("Parcel \"%s\" isn't tracked anymore." % parcel.fullid)

    def do_status(self, line):
        """
        status

        Display status for all of the tracked parcels.
        """
        backends = list(map(get_backend_name, self.enabled_backends))
        self.start_format()
        # XXX cleaning of cached objects may be by start_format()?
        self.objects = []
        for id in self.storage.get('tracking', default=[]):
            # It should be safe to do it here, since all objects in storage
            # are stored with the fullid
            _id, backend_name = id.rsplit('@', 1)
            # If the user use the -b or -e option, do not try to get
            # the status of parcel of not loaded backends
            if backend_name not in backends:
                continue

            try:
                p = self.get_object(id, 'get_parcel_tracking')
            except CallErrors as errors:
                print('Error with parcel', id, file=self.stderr)
                self.bcall_errors_handler(errors)
                continue

            if p is None:
                continue
            self.cached_format(p)

    def do_info(self, id):
        """
        info ID

        Get information about a parcel.
        """
        parcel = self.get_object(id, 'get_parcel_tracking', [])
        if not parcel:
            print('Error: parcel not found', file=self.stderr)
            return 2

        self.start_format()
        self.format(parcel)
        for event in parcel.history:
            self.format(event)



# Copyright(C) 2010-2012 Jeremy Monnet
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .collection import CapCollection
from .base import BaseObject, BoolField, StringField
from .date import DateField


__all__ = ['Book', 'Renew', 'CapBook']


class Book(BaseObject):
    """
    Describes a book.
    """
    name =      StringField('Name of the book')
    author =    StringField('Author of the book')
    location =  StringField('Location')
    date =      DateField('The due date')
    late =      BoolField('Are you late?')


class Renew(BaseObject):
    """
    A renew message.
    """
    message = StringField('Message')


class CapBook(CapCollection):
    """
    Library websites.
    """

    def iter_resources(self, objs, split_path):
        """
        Iter resources. It retuns :func:`iter_books`.
        """
        if Book in objs:
            self._restrict_level(split_path)
            return self.iter_books()

    def iter_books(self):
        """
        Iter books.

        :rtype: iter[:class:`Book`]
        """
        raise NotImplementedError()

    def get_book(self, _id):
        """
        Get a book from an ID.

        :param _id: ID of the book
        :type _id: str
        :rtype: :class:`Book`
        """
        raise NotImplementedError()

    def iter_booked(self):
        raise NotImplementedError()

    def renew_book(self, _id):
        raise NotImplementedError()

    def iter_rented(self):
        raise NotImplementedError()

    def search_books(self, _string):
        raise NotImplementedError()



# Copyright(C) 2015      Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


__all__ = ['Javascript']


from woob.tools.log import getLogger


class Javascript:
    HEADER = """
    function btoa(str) {
        var buffer;

        if (str instanceof Buffer) {
            buffer = str;
        } else {
            buffer = new Buffer(str.toString(), 'binary');
        }

        return buffer.toString('base64');
    }

    function atob(str) {
        return new Buffer(str, 'base64').toString('binary');
    }

    document = {
        createAttribute: null,
        styleSheets: null,
        characterSet: "UTF-8",
        documentElement: {}
    };

    history = {};

    screen = {
        width: 1280,
        height: 800
    };

    var XMLHttpRequest = function() {};
    XMLHttpRequest.prototype.onreadystatechange = function(){};
    XMLHttpRequest.prototype.open = function(){};
    XMLHttpRequest.prototype.setRequestHeader = function(){};
    XMLHttpRequest.prototype.send = function(){};

    /* JS code checks that some PhantomJS globals aren't defined on the
    * global window object; put an empty window object, so that all these
    * tests fail.
    * It then tests the user agent against some known scrappers; just put
    * the default Tor user agent in there.
    */
    window = {
        document: document,
        history: history,
        screen: screen,
        XMLHttpRequest: XMLHttpRequest,

        innerWidth: 1280,
        innerHeight: 800,

        close: function(){}
    };

    navigator = {
        userAgent: "Mozilla/5.0 (X11; Linux x86_64; rv:68.0) Gecko/20100101 Firefox/68.0",
        appName: "Netscape"
    };
    """

    def __init__(self, script, logger=None, domain=""):
        try:
            import execjs
        except ImportError:
            raise ImportError('Please install PyExecJS')

        self.runner = execjs.get()
        self.logger = getLogger('js', logger)

        window_emulator = self.HEADER

        if domain:
            window_emulator += "document.domain = '" + domain + "';"
            window_emulator += """
            if (typeof(location) === "undefined") {
                var location = window.location = {
                    host: document.domain
                };
            }
            """

        self.ctx = self.runner.compile(window_emulator + script)

    def call(self, *args, **kwargs):
        retval = self.ctx.call(*args, **kwargs)

        self.logger.debug('Calling %s%s = %s', args[0], args[1:], retval)

        return retval



# Copyright(C) 2010-2012 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import os
import re
from shutil import which
import subprocess

import requests

from woob.capabilities.radio import CapRadio, Radio
from woob.capabilities.audio import CapAudio, BaseAudio, Playlist, Album
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.media_player import InvalidMediaPlayer, MediaPlayer, MediaPlayerNotFound
from woob.tools.application.formatters.iformatter import PrettyFormatter

__all__ = ['AppRadio']


class RadioListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title')

    def get_title(self, obj):
        return obj.title

    def get_description(self, obj):
        result = ''

        if hasattr(obj, 'description') and not empty(obj.description):
            result += '%-30s' % obj.description

        if hasattr(obj, 'current') and not empty(obj.current):
            if obj.current.who:
                result += ' (Current: %s - %s)' % (obj.current.who, obj.current.what)
            else:
                result += ' (Current: %s)' % obj.current.what
        return result


class SongListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title')

    def get_title(self, obj):
        result = obj.title

        if hasattr(obj, 'author') and not empty(obj.author):
            result += ' (%s)' % obj.author

        return result

    def get_description(self, obj):
        result = ''
        if hasattr(obj, 'description') and not empty(obj.description):
            result += '%-30s' % obj.description

        return result


class AlbumTrackListInfoFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'tracks_list')

    def get_title(self, obj):
        result = obj.title

        if hasattr(obj, 'author') and not empty(obj.author):
            result += ' (%s)' % obj.author

        return result

    def get_description(self, obj):
        result = ''
        for song in obj.tracks_list:
            result += '- %s%-30s%s ' % (self.BOLD, song.title, self.NC)

            if hasattr(song, 'duration') and not empty(song.duration):
                result += '%-10s ' % song.duration
            else:
                result += '%-10s ' % ' '

            result += '(%s)\r\n\t' % (song.id)

        return result


class PlaylistTrackListInfoFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'tracks_list')

    def get_title(self, obj):
        return obj.title

    def get_description(self, obj):
        result = ''
        for song in obj.tracks_list:
            result += '- %s%-30s%s ' % (self.BOLD, song.title, self.NC)

            if hasattr(song, 'author') and not empty(song.author):
                result += '(%-15s) ' % song.author

            if hasattr(song, 'duration') and not empty(song.duration):
                result += '%-10s ' % song.duration
            else:
                result += '%-10s ' % ' '

            result += '(%s)\r\n\t' % (song.id)

        return result


class AppRadio(ReplApplication):
    APPNAME = 'radio'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon\nCopyright(C) YEAR Pierre Maziere'
    DESCRIPTION = "Console application allowing to search for web radio stations, listen to them and get information " \
                  "like the current song."
    SHORT_DESCRIPTION = "search, show or listen to radio stations"
    CAPS = (CapRadio, CapAudio)
    EXTRA_FORMATTERS = {'radio_list': RadioListFormatter,
                        'song_list': SongListFormatter,
                        'album_tracks_list_info': AlbumTrackListInfoFormatter,
                        'playlist_tracks_list_info': PlaylistTrackListInfoFormatter,
                        }

    COMMANDS_FORMATTERS = {'ls': 'radio_list',
                           'playlist': 'radio_list',
                           }

    COLLECTION_OBJECTS = (Radio, BaseAudio, )
    PLAYLIST = []

    def __init__(self, *args, **kwargs):
        super(AppRadio, self).__init__(*args, **kwargs)
        self.player = MediaPlayer(self.logger)

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def complete_download(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_download(self, line):
        """
        download ID [DIRECTORY]

        Download an audio file
        """
        _id, dest = self.parse_command_args(line, 2, 1)

        obj = self.retrieve_obj(_id)

        if obj is None:
            print('No object matches with this id:', _id, file=self.stderr)
            return 3

        if isinstance(obj, BaseAudio):
            streams = [obj]

        else:
            streams = obj.tracks_list

        if len(streams) == 0:
            print('Radio or Audio file not found:', _id, file=self.stderr)
            return 3

        for stream in streams:
            self.download_file(stream, dest)

    def download_file(self, audio, dest):
        _obj = self.get_object(audio.id, 'get_audio', ['url', 'title'])
        if not _obj:
            print('Audio file not found: %s' % audio.id, file=self.stderr)
            return 3

        if not _obj.url:
            print('Error: the direct URL is not available.', file=self.stderr)
            return 4

        audio.url = _obj.url

        def check_exec(executable):
            if which(executable) is None:
                print('Please install "%s"' % executable, file=self.stderr)
                return False
            return True

        def audio_to_file(_audio):
            ext = _audio.ext
            if not ext:
                ext = 'audiofile'
            title = _audio.title if _audio.title else _audio.id
            return '%s.%s' % (re.sub('[?:/]', '-', title), ext)

        if dest is not None and os.path.isdir(dest):
            dest += '/%s' % audio_to_file(audio)

        if dest is None:
            dest = audio_to_file(audio)

        if audio.url.startswith('rtmp'):
            if not check_exec('rtmpdump'):
                return 1
            args = ('rtmpdump', '-e', '-r', audio.url, '-o', dest)
        elif audio.url.startswith('mms'):
            if not check_exec('mimms'):
                return 1
            args = ('mimms', '-r', audio.url, dest)
        else:
            if check_exec('wget'):
                args = ('wget', '-c', audio.url, '-O', dest)
            elif check_exec('curl'):
                args = ('curl', '-C', '-', audio.url, '-o', dest)
            else:
                return 1

        subprocess.call(args)

    def complete_play(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_play(self, line):
        """
        play ID [stream_id]

        Play a radio or a audio file with a found player (optionnaly specify the wanted stream).
        """
        _id, stream_id = self.parse_command_args(line, 2, 1)
        if not _id:
            print('This command takes an argument: %s' % self.get_command_help('play', short=True), file=self.stderr)
            return 2

        try:
            stream_id = int(stream_id)
        except (ValueError, TypeError):
            stream_id = 0

        obj = self.retrieve_obj(_id)

        if obj is None:
            print('No object matches with this id:', _id, file=self.stderr)
            return 3

        if isinstance(obj, Radio):
            try:
                streams = [obj.streams[stream_id]]
            except IndexError:
                print('Stream %d not found' % stream_id, file=self.stderr)
                return 1
        elif isinstance(obj, BaseAudio):
            streams = [obj]

        else:
            streams = obj.tracks_list

        if len(streams) == 0:
            print('Radio or Audio file not found:', _id, file=self.stderr)
            return 3

        try:
            player_name = self.config.get('media_player')
            media_player_args = self.config.get('media_player_args')
            if not player_name:
                self.logger.debug('You can set the media_player key to the player you prefer in the radio '
                                  'configuration file.')

            for stream in streams:
                if isinstance(stream, BaseAudio) and not stream.url:
                    stream = self.get_object(stream.id, 'get_audio')
                else:
                    r = requests.get(stream.url, stream=True, timeout=30)
                    buf = next(r.iter_content(512)).decode('utf-8', 'replace')
                    r.close()
                    playlistFormat = None
                    for line in buf.split("\n"):
                        if playlistFormat is None:
                            if line == "[playlist]":
                                playlistFormat = "pls"
                            elif line == "#EXTM3U":
                                playlistFormat = "m3u"
                            else:
                                break
                        elif playlistFormat == "pls":
                            if line.startswith('File'):
                                stream.url = line.split('=', 1).pop(1).strip()
                                break
                        elif playlistFormat == "m3u":
                            if line[0] != "#":
                                stream.url = line.strip()
                                break

                self.player.play(stream, player_name=player_name, player_args=media_player_args)

        except (InvalidMediaPlayer, MediaPlayerNotFound) as e:
            print('%s\nRadio URL: %s' % (e, stream.url))

    def retrieve_obj(self, _id):
        obj = None
        if self.interactive:
            try:
                obj = self.objects[int(_id) - 1]
                _id = obj.id
            except (IndexError, ValueError):
                pass

        m = CapAudio.get_object_method(_id)
        if m:
            obj = self.get_object(_id, m)

        return obj if obj is not None else self.get_object(_id, 'get_radio')

    def do_playlist(self, line):
        """
        playlist cmd [args]
        playlist add ID [ID2 ID3 ...]
        playlist remove ID [ID2 ID3 ...]
        playlist export [FILENAME]
        playlist display
        """

        if not line:
            print('This command takes an argument: %s' % self.get_command_help('playlist'), file=self.stderr)
            return 2

        cmd, args = self.parse_command_args(line, 2, req_n=1)
        if cmd == "add":
            _ids = args.strip().split(' ')
            for _id in _ids:
                audio = self.get_object(_id, 'get_audio')

                if not audio:
                    print('Audio file not found: %s' % _id, file=self.stderr)
                    return 3

                if not audio.url:
                    print('Error: the direct URL is not available.', file=self.stderr)
                    return 4

                self.PLAYLIST.append(audio)

        elif cmd == "remove":
            _ids = args.strip().split(' ')
            for _id in _ids:

                audio_to_remove = self.get_object(_id, 'get_audio')

                if not audio_to_remove:
                    print('Audio file not found: %s' % _id, file=self.stderr)
                    return 3

                if not audio_to_remove.url:
                    print('Error: the direct URL is not available.', file=self.stderr)
                    return 4

                for audio in self.PLAYLIST:
                    if audio.id == audio_to_remove.id:
                        self.PLAYLIST.remove(audio)
                        break

        elif cmd == "export":
            filename = "playlist.m3u"
            if args:
                filename = args

            file = open(filename, 'w')
            for audio in self.PLAYLIST:
                file.write('%s\r\n' % audio.url)
            file.close()

        elif cmd == "display":
            for audio in self.PLAYLIST:
                self.cached_format(audio)

        else:
            print('Playlist command only support "add", "remove", "display" and "export" arguments.', file=self.stderr)
            return 2

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, _id):
        """
        info ID

        Get information about a radio or an audio file.
        """
        if not _id:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        obj = self.retrieve_obj(_id)

        if isinstance(obj, Album):
            self.set_formatter('album_tracks_list_info')
        elif isinstance(obj, Playlist):
            self.set_formatter('playlist_tracks_list_info')

        if obj is None:
            print('No object matches with this id:', _id, file=self.stderr)
            return 3

        self.format(obj)

    @defaultcount(10)
    def do_search(self, pattern=None):
        """
        search (radio|song|file|album|playlist) PATTERN

        List (radio|song|file|album|playlist) matching a PATTERN.

        If PATTERN is not given, this command will list all the (radio|song|album|playlist).
        """

        if not pattern:
            print('This command takes an argument: %s' % self.get_command_help('search'), file=self.stderr)
            return 2

        cmd, args = self.parse_command_args(pattern, 2, req_n=1)
        if not args:
            args = ""

        self.set_formatter_header('Search pattern: %s' % pattern if pattern else 'All radios')
        self.change_path(['search'])

        if cmd == "radio":
            if 'search' in self.commands_formatters:
                self.set_formatter('radio_list')
            for radio in self.do('iter_radios_search', pattern=args):
                self.add_object(radio)
                self.format(radio)

        elif cmd == "song" or cmd == "file":
            if 'search' in self.commands_formatters:
                self.set_formatter('song_list')
            for audio in self.do('search_audio', pattern=args):
                self.add_object(audio)
                self.format(audio)

        elif cmd == "album":
            if 'search' in self.commands_formatters:
                self.set_formatter('song_list')
            for album in self.do('search_album', pattern=args):
                self.add_object(album)
                self.format(album)

        elif cmd == "playlist":
            if 'search' in self.commands_formatters:
                self.set_formatter('song_list')
            for playlist in self.do('search_playlist', pattern=args):
                self.add_object(playlist)
                self.format(playlist)

        else:
            print('Search command only supports "radio", "song", "file", "album" and "playlist" arguments.', file=self.stderr)
            return 2

    def do_ls(self, line):
        """
        ls

        List radios
        """
        ret = super(AppRadio, self).do_ls(line)
        return ret



# Copyright(C) 2011-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import Capability, BaseObject, NotLoaded, StringField, BoolField, UserError


__all__ = ['PasteNotFound', 'BasePaste', 'CapPaste']


class PasteNotFound(UserError):
    """
    Raised when a paste is not found.
    """

    def __init__(self):
        return super(PasteNotFound, self).__init__("Paste not found")


class BasePaste(BaseObject):
    """
    Represents a pasted text.
    """
    title =         StringField('Title of paste')
    language =      StringField('Language of the paste')
    contents =      StringField('Content of the paste')
    public =        BoolField('Is this paste public?')

    def __init__(self, _id, title=NotLoaded, language=NotLoaded, contents=NotLoaded,
            public=NotLoaded, url=None):
        super(BasePaste, self).__init__(str(_id), url)

        self.title = title
        self.language = language
        self.contents = contents
        self.public = public

    @classmethod
    def id2url(cls, _id):
        """Overloaded in child classes provided by backends."""
        raise NotImplementedError()

    @property
    def page_url(self):
        """
        Get URL to page of this paste.
        """
        return self.id2url(self.id)


class CapPaste(Capability):
    """
    This capability represents the ability for a website backend to store plain text.
    """

    def new_paste(self, *args, **kwargs):
        """
        Get a new paste object for posting it with the backend.
        The parameters should be passed to the object init.

        :rtype: :class:`BasePaste`
        """
        raise NotImplementedError()

    def can_post(self, contents, title=None, public=None, max_age=None):
        """
        Checks if the paste can be pasted by this backend.
        Some properties are considered required (public/private, max_age) while others
        are just bonuses (language).

        contents: Can be used to check encodability, maximum length, etc.
        title: Can be used to check length, allowed characters. Should not be required.
        public: True must be public, False must be private, None do not care.
        max_age: Maximum time to live in seconds.

        A score of 0 means the backend is not suitable.
        A score of 1 means the backend is suitable.
        Higher scores means it is more suitable than others with a lower score.

        :rtype: int
        :returns: score
        """
        raise NotImplementedError()

    def get_paste(self, url):
        """
        Get a Paste from an ID or URL.

        :param _id: the paste id. It can be an ID or a page URL.
        :type _id: str
        :rtype: :class:`BasePaste`
        :raises: :class:`PasteNotFound`
        """
        raise NotImplementedError()

    def post_paste(self, paste, max_age=None):
        """
        Post a paste.

        :param paste: a Paste object
        :type paste: :class:`BasePaste`
        """
        raise NotImplementedError()



# Copyright(C) 2018 Quentin Defenouillere
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import Capability, BaseObject, StringField, UserError


__all__ = [
    'BandInfo', 'BandSearch', 'Albums', 'BandNotFound',
    'Suggestion', 'Favorite',
    'CapBands',
]


class BandSearch(BaseObject):
    """
    Bands search.
    """
    name =                  StringField('Name of band')
    short_description =     StringField('Short description of the band')

    def __init__(self, id='', name=None, short_description=None, url=None):
        super(BandSearch, self).__init__(id, url)
        self.name = name
        self.short_description = short_description


class BandNotFound(UserError):
    """
    Raised when no band is found.
    """
    pass


class BandInfo(BaseObject):
    """
    Information about one specific band.
    """
    name =             StringField('Name of band')
    genre =            StringField('Music genre of the band')
    year =             StringField('Year of creation')
    country =          StringField('Country of origin')
    description =      StringField('Description of the band')

    def __init__(self, name=None, year=None, country=None, genre=None, description=None, url=None):
        super(BandInfo, self).__init__(id, url)
        self.name = name
        self.genre = genre
        self.year = year
        self.description = description
        self.country = country


class Albums(BaseObject):
    """
    Information about one specific band.
    """
    name =             StringField('Album name')
    album_type =       StringField('Type of album')
    year =             StringField('Year of release')
    reviews =          StringField('Album reviews')

    def __init__(self, name=None, album_type=None, year=None, reviews=None, url=None):
        super(Albums, self).__init__(id, url)
        self.name = name
        self.album_type = album_type
        self.year = year
        self.reviews = reviews


class Favorite(BaseObject):
    """
    Fetch your favorite bands.
    """
    name =                  StringField('Name of favorite band')
    band_url =              StringField('URL of the favorite band')
    short_description =     StringField('Short description of the favorite band')

    def __init__(self, id='', name=None, band_url=None, short_description=None):
        super(Favorite, self).__init__(id, name)
        self.name = name
        self.band_url = band_url
        self.short_description = short_description


class Suggestion(BaseObject):
    """
    Band suggestions based on your favorite bands.
    """
    name =                  StringField('Name of suggested band')
    description =           StringField('Band description')
    url =                   StringField('URL of suggested band')

    def __init__(self, id='', name=None, description=None, url=None):
        super(Suggestion, self).__init__(id, url)
        self.name = name
        self.url = url
        self.description = description


class CapBands(Capability):
    """
    Capability to get band information on music websites.
    """

    def iter_band_search(self):
        """
        Look for a band.
        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`BandSearch`]
        """
        raise NotImplementedError()

    def get_info(self):
        """
        Get band info.
        :param band_id: ID of the band
        :rtype: :class:`BandInfo`
        """
        raise NotImplementedError()

    def get_albums(self):
        """
        Get a band's discography.

        :rtype: iter[:class:`Albums`]
        """
        raise NotImplementedError()

    def get_favorites(self):
        """
        Get my favorite bands.

        :rtype: iter[:class:`Favorite`]
        """
        raise NotImplementedError()

    def suggestions(self):
        """
        Get band suggestions according to your favorite bands.

        :rtype: iter[:class:`Suggestion`]
        """
        raise NotImplementedError()



# Copyright(C) 2010-2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


__all__ = ['html2text']


from html2text import HTML2Text


def html2text(html, **options):
    h = HTML2Text()
    defaults = dict(
        unicode_snob=True,
        skip_internal_links=True,
        inline_links=False,
        links_each_paragraph=True,
    )
    defaults.update(options)
    for k, v in defaults.items():
        setattr(h, k, v)
    return str(h.handle(html))



# Copyright(C) 2014-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import datetime

from dateutil.relativedelta import relativedelta
from requests.exceptions import HTTPError
from woob.exceptions import (
    BrowserHTTPError, BrowserHTTPNotFound, BrowserUnavailable,
)


class HTTPNotFound(HTTPError, BrowserHTTPNotFound):
    pass


class ClientError(HTTPError, BrowserHTTPError):
    pass


class ServerError(HTTPError, BrowserHTTPError, BrowserUnavailable):
    pass


class LoggedOut(Exception):
    pass


class BrowserTooManyRequests(BrowserUnavailable):
    """
    Client tries to perform too many requests within a certain timeframe.
    The module should set the next_try if possible, else it is set to 24h.
    """

    def __init__(self, message='', next_try=None):
        super(BrowserTooManyRequests, self).__init__(message)

        if isinstance(next_try, datetime.date) and not isinstance(next_try, datetime.datetime):
            next_try = datetime.datetime.combine(next_try, datetime.datetime.min.time())

        if next_try is None:
            next_try = datetime.datetime.now() + relativedelta(days=1)

        if not isinstance(next_try, datetime.datetime):
            raise TypeError('next_try value should be a datetime.')

        self.next_try = next_try

    def __str__(self):
        return super(BrowserTooManyRequests, self).__str__() or 'Too many requests, next_try set %s' % self.next_try



# Copyright(C) 2019-2020 Célande Adrien
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.rpg import CapRPG, Character, Skill, CharacterClass, CollectableItem
from woob.tools.application.repl import ReplApplication, defaultcount


class AppRPG(ReplApplication):
    APPNAME = 'rpg'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2019-YEAR Célande Adrien'
    CAPS = CapRPG
    DESCRIPTION = 'Console application allowing to list informations from a RPG.'
    SHORT_DESCRIPTION = 'manage RPG data'
    DEFAULT_FORMATTER = 'table'
    COLLECTION_OBJECTS = (Character, Skill, CharacterClass, CollectableItem, )

    def do_characters(self, line):
        """
        characters

        List all characters
        """
        print('do characterS')
        print(f'line: "{line}"')
        return self.do_ls(line)

    def do_character(self, line):
        """
        character ID

        Get data on one character
        """
        character_id, = self.parse_command_args(line, 1, 1)
        self.start_format()
        # cannot use get_object because it can be skipped after a ls
        for c in self.do('get_character', character_id):
            self.format(c)

    @defaultcount(20)
    def do_skills(self, line):
        """
        skills [TYPE]

        List all skills
        """
        print('do skills')
        skill_type, = self.parse_command_args(line, 1, 0)
        print('skill type', skill_type)
        self.start_format()
        for skill in self.do('iter_skills', skill_type):
            self.format(skill)

    def do_skill(self, line):
        """
        skill ID

        Details for one skill
        """
        skill_id, = self.parse_command_args(line, 1, 1)
        self.start_format()
        skill = self.get_object(skill_id, 'get_skill', [])
        self.format(skill)

    @defaultcount(20)
    def do_skill_set(self, line):
        """
        skill_set CHARACTER_ID [TYPE]

        List of skills for a character
        """
        character_id, skill_type = self.parse_command_args(line, 2, 1)
        self.start_format()
        for skill in  self.do('iter_skill_set', character_id, skill_type):
            self.format(skill)

    @defaultcount(20)
    def do_classes(self, line):
        """
        classes

        List all character classes
        """
        self.start_format()
        for character_class in self.do('iter_character_classes'):
            self.format(character_class)

    def do_class(self, line):
        """
        class ID

        Details for one character class
        """
        class_id, = self.parse_command_args(line, 1, 1)
        self.start_format()
        character_class = self.get_object(class_id, 'get_character_class', [])
        self.format(character_class)

    @defaultcount(20)
    def do_items(self, line):
        """
        items

        List all collectable items
        """
        self.start_format()
        for item in self.do('iter_collectable_items'):
            self.format(item)

    def do_item(self, line):
        """
        item ID

        Details for one collectable item
        """
        item_id, = self.parse_command_args(line, 1, 1)
        self.start_format()
        item = self.get_object(item_id, 'get_collectable_item', [])
        self.format(item)



# Copyright(C) 2017  Vincent A
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from shutil import which
from datetime import timedelta
import subprocess

from woob.capabilities.base import NotAvailable
from woob.capabilities.image import Thumbnail
from woob.capabilities.video import BaseVideo
from woob.tools.date import parse_date
from woob.tools.json import json


__all__ = ('video_info',)


def video_info(url):
    """Fetch info about a video using youtube-dl

    :param url: URL of the web page containing the video
    :rtype: :class:`woob.capabilities.video.Video`
    """

    ytdl = which('yt-dlp') or which('youtube-dl')
    if ytdl is None:
        raise Exception('Please install youtube-dl')

    try:
        j = json.loads(subprocess.check_output([ytdl, '-J', url]))
    except subprocess.CalledProcessError:
        return None

    v = BaseVideo(id=url)
    v.title = str(j.get('title')) if j.get('title') else NotAvailable
    v.ext = str(j.get('ext')) if j.get('ext') else NotAvailable
    v.description = str(j.get('description')) if j.get('description') else NotAvailable
    v.url = str(j['url'])
    v.duration = timedelta(seconds=j.get('duration')) if j.get('duration') else NotAvailable
    v.author = str(j.get('uploader')) if j.get('uploader') else NotAvailable
    v.rating = j.get('average_rating') or NotAvailable

    if j.get('thumbnail'):
        v.thumbnail = Thumbnail(str(j['thumbnail']))

    d = j.get('upload_date', j.get('release_date'))
    if d:
        v.date = parse_date(d)

    return v



# Copyright(C) 2010-2013 Romain Bignon
# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, Field, StringField
from woob.tools.capabilities.streaminfo import StreamInfo


__all__ = ['Radio', 'CapRadio']


class Radio(BaseObject):
    """
    Radio object.
    """
    title =         StringField('Title of radio')
    description =   StringField('Description of radio')
    current =       Field('Current emission', StreamInfo)
    streams =       Field('List of streams', list)


class CapRadio(Capability):
    """
    Capability of radio websites.
    """

    def iter_radios_search(self, pattern):
        """
        Search a radio.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`Radio`]
        """
        raise NotImplementedError()

    def get_radio(self, id):
        """
        Get a radio from an ID.

        :param id: ID of radio
        :type id: str
        :rtype: :class:`Radio`
        """
        raise NotImplementedError()



# Copyright(C) 2010-2013  Christophe Benz, Julien Hebert
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from collections import OrderedDict
from pathlib import Path
import os
import sys
import subprocess
import warnings

try:
    import tty
    import termios
except ImportError:
    PROMPT = '--Press return to continue--'

    def readch():
        return sys.stdin.readline()
else:
    PROMPT = '--Press a key to continue--'

    def readch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)

        tty.setcbreak(fd)
        try:
            c = sys.stdin.read(1)
            # XXX do not read magic number
            if c == '\x03':
                raise KeyboardInterrupt()
            return c
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

from woob.capabilities.base import BaseObject
from woob.tools.application.pretty import colored, NC, BOLD
from woob.tools.misc import guess_encoding, classproperty


__all__ = ['IFormatter', 'MandatoryFieldsNotFound']


class MandatoryFieldsNotFound(Exception):
    def __init__(self, missing_fields):
        super().__init__('Mandatory fields not found: %s.' % ', '.join(missing_fields))


class IFormatter:
    # Tuple of fields mandatory to not crash
    MANDATORY_FIELDS = None
    # Tuple of displayed field. Set to None if all available fields are
    # displayed
    DISPLAYED_FIELDS = None

    @classproperty
    def BOLD(self):
        warnings.warn(
            'Use woob.tools.application.pretty.BOLD instead.\n'
            'That\'s also better to use woob.tools.application.pretty.colored.',
            DeprecationWarning,
            stacklevel=2
        )
        return BOLD

    @classproperty
    def NC(self):
        warnings.warn(
            'Use woob.tools.application.pretty.NC instead.\n'
            'That\'s also better to use woob.tools.application.pretty.colored.',
            DeprecationWarning,
            stacklevel=2
        )
        return NC

    def colored(self, string, color, attrs=None, on_color=None):
        if (
            self.outfile != sys.stdout
            or not self.outfile.isatty()
            or os.getenv("NO_COLOR") is not None
        ):
            return string

        if isinstance(attrs, str):
            attrs = [attrs]
        return colored(string, color, on_color=on_color, attrs=attrs)

    def __init__(self, display_keys=True, display_header=True, outfile=None):
        self.display_keys = display_keys
        self.display_header = display_header
        self.interactive = False
        self.print_lines = 0
        self.termrows = 0
        self.termcols = None
        if outfile is None:
            outfile = sys.stdout
        self.outfile = outfile
        # XXX if stdin is not a tty, it seems that the command fails.

        if sys.stdout.isatty() and sys.stdin.isatty():
            if sys.platform == 'win32':
                from ctypes import windll, create_string_buffer

                h = windll.kernel32.GetStdHandle(-12)
                csbi = create_string_buffer(22)
                res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)

                if res:
                    import struct
                    (bufx, bufy, curx, cury, wattr,
                     left, top, right, bottom, maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                    self.termrows = right - left + 1
                    self.termcols = bottom - top + 1
                else:
                    self.termrows = 80  # can't determine actual size - return default values
                    self.termcols = 80
            else:
                stty_output = subprocess.check_output(["stty", "size"]).split()
                self.termrows = int(stty_output[0])
                self.termcols = int(stty_output[1])

    def output(self, formatted):
        if isinstance(self.outfile, (str, Path)):
            encoding = guess_encoding(sys.stdout)
            with open(self.outfile, "a+", encoding=encoding, errors='replace') as outfile:
                outfile.write(formatted + os.linesep)

        else:
            for line in formatted.split('\n'):
                if self.termrows and (self.print_lines + 1) >= self.termrows:
                    self.outfile.write(PROMPT)
                    self.outfile.flush()
                    readch()
                    self.outfile.write('\b \b' * len(PROMPT))
                    self.print_lines = 0

                plen = len(line.replace(BOLD, '').replace(NC, ''))

                print(line, file=self.outfile)

                if self.termcols:
                    self.print_lines += int(plen/self.termcols) + 1
                else:
                    self.print_lines += 1

    def start_format(self, **kwargs):
        pass

    def flush(self):
        pass

    def format(self, obj, selected_fields=None, alias=None):
        """
        Format an object to be human-readable.
        An object has fields which can be selected.

        :param obj: object to format
        :type obj: BaseObject or dict
        :param selected_fields: fields to display. If None, all fields are selected
        :type selected_fields: tuple
        :param alias: an alias to use instead of the object's ID
        :type alias: str
        """
        if isinstance(obj, BaseObject):
            if selected_fields:  # can be an empty list (nothing to do), or None (return all fields)
                obj = obj.copy()
                for name, value in list(obj.iter_fields()):
                    if name not in selected_fields:
                        delattr(obj, name)

            if self.MANDATORY_FIELDS:
                missing_fields = set(self.MANDATORY_FIELDS) - set([name for name, value in obj.iter_fields()])
                if missing_fields:
                    raise MandatoryFieldsNotFound(missing_fields)

            formatted = self.format_obj(obj, alias)
        else:
            try:
                obj = OrderedDict(obj)
            except ValueError:
                raise TypeError('Please give a BaseObject or a dict')

            if selected_fields:
                obj = obj.copy()
                for name, value in list(obj.items()):
                    if name not in selected_fields:
                        obj.pop(name)

            if self.MANDATORY_FIELDS:
                missing_fields = set(self.MANDATORY_FIELDS) - set(obj)
                if missing_fields:
                    raise MandatoryFieldsNotFound(missing_fields)

            formatted = self.format_dict(obj)

        if formatted:
            self.output(formatted)
        return formatted

    def format_obj(self, obj, alias=None):
        """
        Format an object to be human-readable.
        Called by format().
        This method has to be overridden in child classes.

        :param obj: object to format
        :type obj: BaseObject
        :rtype: str
        """
        return self.format_dict(obj.to_dict())

    def format_dict(self, obj):
        """
        Format a dict to be human-readable.

        :param obj: dict to format
        :type obj: dict
        :rtype: str
        """
        raise NotImplementedError()

    def format_collection(self, collection, only):
        """
        Format a collection to be human-readable.

        :param collection: collection to format
        :type collection: BaseCollection
        :rtype: str
        """
        if only is False or collection.basename in only:
            if collection.basename and collection.title:
                self.output('%s~ (%s) %s (%s)%s' %
                     (self.BOLD, collection.basename, collection.title, collection.backend, self.NC))
            else:
                self.output('%s~ (%s) (%s)%s' %
                     (self.BOLD, collection.basename, collection.backend, self.NC))


class PrettyFormatter(IFormatter):
    def format_obj(self, obj, alias):
        title = self.get_title(obj)
        desc = self.get_description(obj)

        if alias is not None:
            result = '%s %s %s (%s)' % (self.colored('%2s' % alias, 'red', 'bold'),
                                         self.colored('—', 'cyan', 'bold'),
                                         self.colored(title, 'yellow', 'bold'),
                                         self.colored(obj.backend, 'blue', 'bold'))
        else:
            result = '%s %s %s' % (self.colored(obj.fullid, 'red', 'bold'),
                                    self.colored('—', 'cyan', 'bold'),
                                    self.colored(title, 'yellow', 'bold'))

        if desc is not None:
            result += '%s\t%s' % (os.linesep, self.colored(desc, 'white'))

        return result

    def get_title(self, obj):
        raise NotImplementedError()

    def get_description(self, obj):
        return None



# Copyright(C) 2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import datetime
import re
import pycountry
from typing import Any
from collections.abc import Iterator
from decimal import Decimal, InvalidOperation
from itertools import islice
from numbers import Number
from urllib.parse import parse_qs, urlparse

from dateutil.parser import parse as parse_date
from dateutil.tz import gettz
from lxml.etree import ElementBase as LXMLElement

from woob.browser.url import URL
from woob.capabilities.base import Currency as BaseCurrency
from woob.capabilities.base import empty
from woob.tools.misc import clean_text

from .base import _NO_DEFAULT, Filter, FilterError, ItemNotFound, _Filter, debug

__all__ = [
    'Filter', 'FilterError', 'RegexpError', 'FormatError',
    'AsyncLoad', 'Async', 'Base', 'Decode', 'Env', 'RawText',
    'CleanText', 'Lower', 'Upper', 'Title', 'Currency', 'NumberFormatError',
    'CleanDecimal', 'Slugify', 'Type', 'Field', 'Regexp', 'Map', 'MapIn',
    'DateTime', 'FromTimestamp', 'Date', 'DateGuesser', 'Time', 'Duration',
    'MultiFilter', 'CombineDate', 'Format', 'BrowserURL', 'Join', 'MultiJoin',
    'Eval', 'QueryValue', 'Coalesce', 'CountryCode',
]


class RegexpError(FilterError):
    pass


class FormatError(FilterError):
    pass


class AsyncLoad(Filter):
    """Load a page asynchronously for later use.

    Often used in combination with :class:`Async` filter.
    """

    def __call__(self, item):
        link = self.select(self.selector, item)
        return item.page.browser.async_open(link) if link else None


class Async(Filter):
    """Selector that uses another page fetched earlier.

    Often used in combination with :class:`AsyncLoad` filter.
    Requires that the other page's URL is matched with a Page by the Browser.

    Example::

        class item(ItemElement):
            load_details = Field('url') & AsyncLoad

            obj_description = Async('details') & CleanText('//h3')
    """

    def __init__(self, name, selector=None):
        super(Async, self).__init__()
        self.selector = selector
        self.name = name

    def __and__(self, o):
        if isinstance(o, type) and issubclass(o, _Filter):
            o = o()
        self.selector = o
        return self

    def __call__(self, item):
        if item.loaders[self.name] is None:
            return None

        return self.select(self.selector, self.loaded_page(item).doc)

    def filter(self, *args):
        raise AttributeError()

    def loaded_page(self, item):
        result = item.loaders[self.name].result()
        assert result.page is not None, 'The loaded url %s hasn\'t been matched by an URL object' % result.url
        return result.page


class Base(Filter):
    """
    Change the base element used in filters.

    >>> Base(Env('header'), CleanText('./h1'))  # doctest: +SKIP
    """

    def __call__(self, item):
        base = self.select(self.base, item)
        if isinstance(base, list):
            assert len(base) == 1, 'If using a list, there must be one element only'
            base = base[0]
        return self.select(self.selector, base)

    def __init__(self, base, selector=None, default=_NO_DEFAULT):
        super(Base, self).__init__(selector, default)
        self.base = base


class Decode(Filter):
    """
    Filter that aims to decode urlencoded strings

    >>> Decode(Env('_id'))  # doctest: +ELLIPSIS
    <woob.browser.filters.standard.Decode object at 0x...>
    >>> from .html import Link
    >>> Decode(Link('./a'))  # doctest: +ELLIPSIS
    <woob.browser.filters.standard.Decode object at 0x...>
    """

    def __call__(self, item):
        self.encoding = item.page.ENCODING if item.page.ENCODING else 'utf-8'
        return self.filter(self.select(self.selector, item))

    @debug()
    def filter(self, txt):
        try:
            try:
                from urllib.parse import unquote
                txt = unquote(txt, self.encoding)
            except ImportError:
                from urllib import unquote
                txt = unquote(txt.encode('ascii')).decode(self.encoding)
        except (UnicodeDecodeError, UnicodeEncodeError):
            pass

        return txt


class Env(_Filter):
    """
    Filter to get environment value of the item.

    It is used for example to get page parameters, or when there is a parse()
    method on ItemElement.
    """

    def __init__(self, name, default=_NO_DEFAULT):
        super(Env, self).__init__(default)
        self.name = name

    def __call__(self, item):
        try:
            return item.env[self.name]
        except KeyError:
            return self.default_or_raise(ItemNotFound('Environment variable %s not found' % self.name))


class RawText(Filter):
    """Get raw text from an element.

    Unlike :class:`CleanText`, whitespace is kept as is.
    """

    def __init__(self, selector=None, children=False, default=_NO_DEFAULT):
        """
        :param children: whether to get text from children elements of the select elements
        :type children: bool
        """

        super(RawText, self).__init__(selector, default=default)
        self.children = children

    @debug()
    def filter(self, el):
        if isinstance(el, (tuple, list)):
            return ' '.join([self.filter(e) for e in el])

        if self.children:
            text = el.text_content()
        else:
            text = el.text

        if text is None:
            result = self.default
        else:
            result = str(text)

        return result


class CleanText(Filter):
    """
    Get a cleaned text from an element.

    It first replaces all tabs and multiple spaces
    (including newlines if ``newlines`` is True)
    to one space and strips the result string.

    The result is coerced into str, and optionally normalized
    according to the ``normalize`` argument.

    Then it replaces all symbols given in the ``symbols`` argument.

    >>> CleanText().filter('coucou ') == u'coucou'
    True
    >>> CleanText().filter(u'coucou\xa0coucou') == u'coucou coucou'
    True
    >>> CleanText(newlines=True).filter(u'coucou\\r\\n coucou ') == u'coucou coucou'
    True
    >>> CleanText(newlines=False).filter(u'coucou\\r\\n coucou ') == u'coucou\\ncoucou'
    True
    """

    def __init__(self, selector=None, symbols='', replace=[], children=True, newlines=True, transliterate=False, normalize='NFC', **kwargs):
        """
        :param symbols: list of strings to remove from text
        :type symbols: list
        :param replace: optional pairs of text replacements to perform
        :type replace: list[tuple[str, str]]
        :param children: whether to get text from children elements of the select elements
        :type children: bool
        :param newlines: if True, newlines will be converted to space too
        :type newlines: bool
        :param normalize: Unicode normalization to perform
        :type normalize: str or None
        :param transliterate: Transliterates unicode characters into ASCII characters
        :type transliterate: bool
        """

        super(CleanText, self).__init__(selector, **kwargs)
        self.symbols = symbols
        self.toreplace = replace
        self.children = children
        self.newlines = newlines
        self.normalize = normalize
        self.transliterate = transliterate

    @debug()
    def filter(self, txt):
        if txt is None:
            return self.default_or_raise(FilterError('The text cannot be None'))
        elif isinstance(txt, int):
            txt = str(txt)
        elif isinstance(txt, (tuple, list)):
            txt = ' '.join(
                self.clean(item, newlines=self.newlines, children=self.children)
                for item in txt
            )

        txt = self.clean(txt, self.children, self.newlines, self.normalize, self.transliterate)
        txt = self.remove(txt, self.symbols)
        txt = self.replace(txt, self.toreplace)
        return txt

    @classmethod
    def clean(cls, txt, children=True, newlines=True, normalize='NFC', transliterate=False):
        """
        Cleans the text. The children argument is ignored with Selenium.
        """
        if isinstance(txt, LXMLElement):
            if children:
                txt = list(txt.itertext())
            else:
                txt = list(txt.xpath('./text()'))
            txt = ' '.join(txt)  # 'foo   bar '
        elif not isinstance(txt, str):
            txt = ' '.join(txt.itertext())

        return clean_text(
            txt,
            remove_newlines=newlines,
            normalize=normalize,
            transliterate=transliterate,
        )

    @classmethod
    def remove(cls, txt, symbols):
        for symbol in symbols:
            txt = txt.replace(symbol, '')
        return txt.strip()

    @classmethod
    def replace(cls, txt, replace):
        for before, after in replace:
            txt = txt.replace(before, after)
        return txt


class Lower(CleanText):
    """Extract text with :class:`CleanText` and convert to lower-case."""

    @debug()
    def filter(self, txt):
        txt = super(Lower, self).filter(txt)
        return txt.lower()


class Upper(CleanText):
    """Extract text with :class:`CleanText` and convert to upper-case."""

    @debug()
    def filter(self, txt):
        txt = super(Upper, self).filter(txt)
        return txt.upper()


class Title(CleanText):
    """Extract text with :class:`CleanText` and apply title() to it."""

    @debug()
    def filter(self, txt):
        txt = super(Title, self).filter(txt)
        return txt.title()


class Currency(CleanText):
    @debug()
    def filter(self, txt):
        txt = super(Currency, self).filter(txt)
        if empty(txt):
            return self.default_or_raise(FormatError("Unable to parse %r" % txt))
        return BaseCurrency.get_currency(txt)


class CountryCode(CleanText):
    """
    Filter to get the country ISO 3166-1 alpha-2 code from the country name
    """
    @debug()
    def filter(self, txt: str) -> Any:
        """Get the country code from the name of the country

        :param txt: Country name
        :type txt: str
        :raises FormatError: if the Country name is not found

        >>> CountryCode().filter('france')
        'fr'
        >>> CountryCode(default= 'll').filter('Greez')
        'll'
        """
        txt = super().filter(txt)
        if empty(txt):
            return self.default_or_raise(FormatError("Unable to parse %r" % txt))
        try:
            alpha_2_code = pycountry.countries.lookup(txt).alpha_2.lower()
        except LookupError:
            # Country not found
            return self.default_or_raise(FormatError('Country not recognized: %r' % txt))

        return alpha_2_code


class NumberFormatError(FormatError, InvalidOperation):
    pass


class CleanDecimal(CleanText):
    """
    Get a cleaned Decimal value from an element.

    `replace_dots` is False by default. A dot is interpreted as a decimal separator.

    If `replace_dots` is set to True, we remove all the dots. The ',' is used as decimal
    separator (often useful for French values)

    If `replace_dots` is a tuple, the first element will be used as the thousands separator,
    and the second as the decimal separator.

    See https://en.wikipedia.org/wiki/Thousands_separator#Examples_of_use

    For example, for the UK style (as in 1,234,567.89):

    >>> CleanDecimal('./td[1]', replace_dots=(',', '.'))  # doctest: +SKIP
    """

    def __init__(self, selector=None, replace_dots=False, sign=None, legacy=True, default=_NO_DEFAULT):
        """
        :param sign: function accepting the text as param and returning the sign
        """

        super(CleanDecimal, self).__init__(selector, default=default)
        self.replace_dots = replace_dots
        self.sign = sign
        self.legacy = legacy
        if not legacy:
            thousands_sep, decimal_sep = self.replace_dots
            self.matching = re.compile(r'([+-]?)\s*(\d[\d%s%s]*|%s\d+)' % tuple(map(re.escape, (thousands_sep, decimal_sep, decimal_sep))))
            self.thousand_check = re.compile(r'^[+-]?\d{1,3}(%s\d{3})*(%s\d*)?$' % tuple(map(re.escape, (thousands_sep, decimal_sep))))
            self.is_scientific_notation = re.compile(r'([+-]?)(\d+(?:[.,]\d*)?[eE][+-]?\d+)')

    @debug()
    def filter(self, text):
        if type(text) in (float, int):
            text = str(text)

        if empty(text):
            return self.default_or_raise(FormatError('Unable to parse %r' % text))

        original_text = text = super(CleanDecimal, self).filter(text)

        text = text.replace('\u2212', '-')

        if self.legacy:
            if self.replace_dots:
                if type(self.replace_dots) is tuple:
                    thousands_sep, decimal_sep = self.replace_dots
                else:
                    thousands_sep, decimal_sep = '.', ','
                text = text.replace(thousands_sep, '').replace(decimal_sep, '.')

            text = re.sub(r'[^\d\-\.]', '', text)
        else:
            thousands_sep, decimal_sep = self.replace_dots

            matches = self.matching.findall(text)
            if not matches:
                return self.default_or_raise(NumberFormatError('There is no number to parse'))
            elif len(matches) > 1:
                matches = self.is_scientific_notation.findall(text)
                if not matches:
                    return self.default_or_raise(NumberFormatError('There should be exactly one number to parse'))

            text = '%s%s' % (matches[0][0], matches[0][1].strip())

            if thousands_sep and thousands_sep in text and not self.thousand_check.match(text):
                return self.default_or_raise(NumberFormatError('Thousands separator is misplaced in %r' % text))

            text = text.replace(thousands_sep, '').replace(decimal_sep, '.')

        try:
            v = Decimal(text)
        except InvalidOperation as e:
            return self.default_or_raise(NumberFormatError(e))
        else:
            if self.sign is not None:
                if callable(self.sign):
                    v *= self.sign(original_text)
                elif self.sign == '+':
                    return abs(v)
                elif self.sign == '-':
                    return -abs(v)
                else:
                    raise TypeError("'sign' should be a callable or a sign string")
            return v

    @classmethod
    def US(cls, *args, **kwargs):
        kwargs['legacy'] = False
        kwargs['replace_dots'] = (',', '.')
        return cls(*args, **kwargs)

    @classmethod
    def French(cls, *args, **kwargs):
        kwargs['legacy'] = False
        kwargs['replace_dots'] = (' ', ',')
        return cls(*args, **kwargs)

    @classmethod
    def SI(cls, *args, **kwargs):
        kwargs['legacy'] = False
        kwargs['replace_dots'] = (' ', '.')
        return cls(*args, **kwargs)

    @classmethod
    def Italian(cls, *args, **kwargs):
        kwargs['legacy'] = False
        kwargs['replace_dots'] = ('.', ',')
        return cls(*args, **kwargs)


class Slugify(Filter):
    @debug()
    def filter(self, label):
        label = re.sub(r'[^A-Za-z0-9]', ' ', label.lower()).strip()
        label = re.sub(r'\s+', '-', label)
        return label


class Type(Filter):
    """
    Get a cleaned value of any type from an element text.
    The type_func can be any callable (class, function, etc.).
    By default an empty string will not be parsed but it can be changed
    by specifying minlen=False. Otherwise, a minimal length can be specified.

    >>> Type(CleanText('./td[1]'), type=int)  # doctest: +SKIP

    >>> Type(type=int).filter(42)
    42
    >>> Type(type=int).filter('42')
    42
    >>> Type(type=int, default='NaN').filter('')
    'NaN'
    >>> Type(type=list, minlen=False, default=list('ab')).filter('')
    []
    >>> Type(type=list, minlen=0, default=list('ab')).filter('')
    ['a', 'b']
    """

    def __init__(self, selector=None, type=None, minlen=0, default=_NO_DEFAULT):
        super(Type, self).__init__(selector, default=default)
        self.type_func = type
        self.minlen = minlen

    @debug()
    def filter(self, txt):
        if isinstance(txt, self.type_func):
            return txt
        if empty(txt):
            return self.default_or_raise(FormatError('Unable to parse %r' % txt))
        if self.minlen is not False and len(txt) <= self.minlen:
            return self.default_or_raise(FormatError('Unable to parse %r' % txt))
        try:
            return self.type_func(txt)
        except ValueError as e:
            return self.default_or_raise(FormatError('Unable to parse %r: %s' % (txt, e)))


class Field(_Filter):
    """
    Get the attribute of object.

    Example::

        obj_foo = CleanText('//h1')
        obj_bar = Field('foo')

    will make "bar" field equal to "foo" field.
    """

    def __init__(self, name):
        super(Field, self).__init__()
        self.name = name

    def __call__(self, item):
        return item.use_selector(getattr(item, 'obj_%s' % self.name), key=self._key)


# Based on nth from https://docs.python.org/2/library/itertools.html
def nth(iterable, n, default=None):
    "Returns the nth item or a default value, n can be negative, or '*' for all"
    if n == '*':
        return iterable
    if n < 0:
        iterable = reversed(list(iterable))
        n = abs(n) - 1
    return next(islice(iterable, n, None), default)


def ordinal(n):
    "To have some readable debug information: '*' => all, 0 => 1st, 1 => 2nd..."
    if n == '*':
        return 'all'
    i = abs(n)
    n = n - 1 if n < 0 else n + 1
    return str(n) + ('th' if i > 2 else ['st', 'nd', 'rd'][i])


class Regexp(Filter):
    r"""
    Apply a regex.

    >>> from lxml.html import etree
    >>> doc = etree.fromstring('<html><body><p>Date: <span>13/08/1988</span></p></body></html>')
    >>> Regexp(CleanText('//p'), r'Date: (\d+)/(\d+)/(\d+)', '\\3-\\2-\\1')(doc) == u'1988-08-13'
    True

    >>> (Regexp(CleanText('//body'), r'(\d+)', nth=1))(doc) == u'08'
    True
    >>> (Regexp(CleanText('//body'), r'(\d+)', nth=-1))(doc) == u'1988'
    True
    >>> (Regexp(CleanText('//body'), r'(\d+)', template='[\\1]', nth='*'))(doc) == [u'[13]', u'[08]', u'[1988]']
    True
    >>> (Regexp(CleanText('//body'), r'Date:.*'))(doc) == u'Date: 13/08/1988'
    True
    >>> (Regexp(CleanText('//body'), r'^(?!Date:).*', default=None))(doc)
    >>>
    """

    def __init__(self, selector=None, pattern=None, template=None, nth=0, flags=0, default=_NO_DEFAULT):
        super(Regexp, self).__init__(selector, default=default)
        if pattern is None:
            raise FilterError('Missing pattern parameter')
        self.pattern = pattern

        self._regex = re.compile(pattern, flags)
        self.template = template
        self.nth = nth

    def expand(self, m):
        if self.template is None:
            try:
                return next(g for g in m.groups() if g is not None)
            except StopIteration:
                return m.string
        return self.template(m) if callable(self.template) else m.expand(self.template)

    @debug()
    def filter(self, txt):
        """
        :raises: :class:`RegexpError` if `pattern` was not found
        """

        if isinstance(txt, (tuple, list)):
            txt = ' '.join([t.strip() for t in txt.itertext()])

        m = None
        try:
            m = self._regex.search(txt) if self.nth == 0 else \
                nth(self._regex.finditer(txt), self.nth)
        except TypeError:
            msg = '%r is not a string or bytes-like object' % txt
            return self.default_or_raise(RegexpError(msg))

        if not m:
            if len(txt) > 1024:
                txt = txt[:1021] + "..."
            msg = 'Unable to find %s %s in %r' % (ordinal(self.nth), self.pattern, txt)
            return self.default_or_raise(RegexpError(msg))

        if isinstance(m, Iterator):
            return list(map(self.expand, m))

        return self.expand(m)


class Map(Filter):
    """Map selected value to another value using a dict.

    Example::

        TYPES = {
            'Concert': CATEGORIES.CONCERT,
            'Cinéma': CATEGORIES.CINE,
        }

        obj_type = Map(CleanText('./li'), TYPES)
    """

    def __init__(self, selector, map_dict, default=_NO_DEFAULT):
        """
        :param selector: key from `map_dict` to use
        """

        super(Map, self).__init__(selector, default=default)
        self.map_dict = map_dict

    @debug()
    def filter(self, txt):
        """
        :raises: :class:`ItemNotFound` if key does not exist in dict
        """

        try:
            return self.map_dict[txt]
        except KeyError:
            return self.default_or_raise(ItemNotFound('Unable to handle %r on %r' % (txt, self.map_dict)))


class MapIn(Filter):
    """
    Map the pattern of a selected value to another value using a dict.
    """

    def __init__(self, selector, map_dict, default=_NO_DEFAULT):
        """
        :param selector: key from `map_dict` to use
        """
        super(MapIn, self).__init__(selector, default=default)
        self.map_dict = map_dict

    @debug()
    def filter(self, txt):
        """
        :raises: :class:`ItemNotFound` if key pattern does not exist in dict
        """
        if txt is not None:
            for key in self.map_dict:
                if key in txt:
                    return self.map_dict[key]

        return self.default_or_raise(ItemNotFound('Unable to handle %r on %r' % (txt, self.map_dict)))


class DateTime(Filter):
    """Parse date and time."""

    def __init__(self, selector=None, default=_NO_DEFAULT, translations=None,
                 parse_func=parse_date, strict=True, tzinfo=None, **kwargs):
        """
        :param dayfirst: if True, the day is the first element in the string to parse
        :type dayfirst: bool
        :param parse_func: the function to use for parsing the datetime
        :param translations: string replacements from site locale to English
        :type translations: list[tuple[str, str]]
        :param tzinfo: timezone to set if none was parsed
        :type tzinfo: :class:`datetime.tzinfo`
        """

        super(DateTime, self).__init__(selector, default=default)
        self.kwargs = kwargs
        self.translations = translations
        self.parse_func = parse_func
        self.strict = strict
        if isinstance(tzinfo, str):
            tzinfo = gettz(tzinfo)
        self.tzinfo = tzinfo

    _default_date_1 = datetime.datetime(2100, 10, 10, 1, 1, 1)
    _default_date_2 = datetime.datetime(2120, 12, 12, 2, 2, 2)

    @debug()
    def filter(self, txt):
        if empty(txt) or txt == '':
            return self.default_or_raise(FormatError('Unable to parse %r' % txt))
        try:
            if self.translations:
                for search, repl in self.translations:
                    txt = search.sub(repl, txt)
            if self.strict:
                parse1 = self.parse_func(txt, default=self._default_date_1, **self.kwargs)
                parse2 = self.parse_func(txt, default=self._default_date_2, **self.kwargs)
                if parse1 != parse2:
                    raise FilterError('Date is not complete')
            else:
                parse1 = self.parse_func(txt, **self.kwargs)

            if parse1.tzinfo is None and self.tzinfo:
                parse1 = parse1.replace(tzinfo=self.tzinfo)

            return parse1
        except (ValueError, TypeError) as e:
            return self.default_or_raise(FormatError('Unable to parse %r: %s' % (txt, e)))


class FromTimestamp(Filter):
    """Parse a timestamp into a datetime."""

    def __init__(self, selector, millis=False, tz=None, default=_NO_DEFAULT):
        super(FromTimestamp, self).__init__(selector, default=default)
        self.millis = millis
        self.tz = tz

    @debug()
    def filter(self, txt):
        try:
            ts = float(txt)
        except (TypeError, ValueError) as exc:
            return self.default_or_raise(FormatError('Unable to parse %r: %s' % (txt, exc)))

        if self.millis:
            ts /= 1000

        try:
            return datetime.datetime.fromtimestamp(ts, tz=self.tz)
        except TypeError as exc:
            return self.default_or_raise(FormatError('Unable to parse %r: %s' % (txt, exc)))


class Date(DateTime):
    """Parse date."""

    def __init__(self, selector=None, default=_NO_DEFAULT, translations=None, parse_func=parse_date, strict=True, **kwargs):
        super(Date, self).__init__(selector, default=default, translations=translations, parse_func=parse_func, strict=strict, **kwargs)

    _default_date_1 = datetime.datetime(2100, 10, 10, 1, 1, 1)
    _default_date_2 = datetime.datetime(2120, 12, 12, 1, 1, 1)

    @debug()
    def filter(self, txt):
        datetime = super(Date, self).filter(txt)
        if hasattr(datetime, 'date'):
            return datetime.date()
        else:
            return datetime


class DateGuesser(Filter):
    def __init__(self, selector, date_guesser, **kwargs):
        super(DateGuesser, self).__init__(selector)
        self.date_guesser = date_guesser
        self.kwargs = kwargs

    def __call__(self, item):
        values = self.select(self.selector, item)
        date_guesser = self.date_guesser
        # In case Env() is used to kive date_guesser.
        if isinstance(date_guesser, _Filter):
            date_guesser = self.select(date_guesser, item)

        if isinstance(values, str):
            values = re.split('[/-]', values)
        if len(values) == 2:
            day, month = map(int, values)
        else:
            raise FormatError('Unable to take (day, month) tuple from %r' % values)
        return date_guesser.guess_date(day, month, **self.kwargs)


class Time(Filter):
    """Parse time."""

    klass: type = datetime.time
    _regexp = re.compile(r'(?P<hh>\d+)[:h]?(?P<mm>\d+)([:m](?P<ss>\d+))?')
    kwargs = {'hour': 'hh', 'minute': 'mm', 'second': 'ss'}

    def __init__(self, selector=None, default=_NO_DEFAULT):
        super(Time, self).__init__(selector, default=default)

    @debug()
    def filter(self, txt):
        m = self._regexp.search(txt)
        if m:
            kwargs = {}
            for key, index in self.kwargs.items():
                kwargs[key] = int(m.groupdict()[index] or 0)
            return self.klass(**kwargs)

        return self.default_or_raise(FormatError('Unable to find time in %r' % txt))


class Duration(Time):
    """Parse a duration as timedelta."""

    klass: type = datetime.timedelta
    _regexp = re.compile(r'((?P<hh>\d+)[:;])?(?P<mm>\d+)[;:](?P<ss>\d+)')
    kwargs = {'hours': 'hh', 'minutes': 'mm', 'seconds': 'ss'}


class MultiFilter(Filter):
    def __init__(self, *args, **kwargs):
        default = kwargs.pop('default', _NO_DEFAULT)
        super(MultiFilter, self).__init__(args, default)

    def __call__(self, item):
        values = [self.select(selector, item) for selector in self.selector]
        return self.filter(tuple(values))

    def filter(self, values):
        raise NotImplementedError()


class CombineDate(MultiFilter):
    """Combine separate Date and Time filters into a single datetime."""

    def __init__(self, date, time):
        """
        :type date: filter
        :type time: filter
        """
        super(CombineDate, self).__init__(date, time)

    @debug()
    def filter(self, values):
        return datetime.datetime.combine(values[0], values[1])


class Format(MultiFilter):
    """Combine multiple filters with string-format.

    Example::

        obj_title = Format('%s (%s)', CleanText('//h1'), CleanText('//h2'))

    will concatenate the text from all ``<h1>`` and all ``<h2>`` (but put
    the latter between parentheses).
    """

    def __init__(self, fmt, *args):
        """
        :param fmt: string format suitable for "%"-formatting
        :type fmt: str
        :param args: other filters to insert in `fmt` string.
                     There should be as many args as there are "%" in `fmt`.
        """
        super(Format, self).__init__(*args)
        self.fmt = fmt

    @debug()
    def filter(self, values):
        return self.fmt % values


class BrowserURL(MultiFilter):
    r""" Format URL using names in parent Browser

    This filter allows to format URL using an URL defined
    in browser instance of this page.

    .. code-block:: python

    class MyBrowser:

        mypage = URL('(?P<category>\w+)/(?P<id>\w+)')


    class OnePage(Page):
        class item(ItemElement):
            obj_myfield = BrowserURL('mypage', id=Dict('id'), category=Dict('category'))

    """
    def __init__(self, url_name, **kwargs):
        super(BrowserURL, self).__init__(*kwargs.values())
        self.url_name = url_name
        self.keys = list(kwargs.keys())

    def __call__(self, item):
        values = super(BrowserURL, self).__call__(item)
        url = getattr(item.page.browser, self.url_name)
        assert isinstance(url, URL), "%s.%s must be an URL object" % (type(item.page.browser).__name__, self.url_name)
        return url.build(**dict(zip(self.keys, values)))

    @debug()
    def filter(self, values):
        return values


class Join(Filter):
    """
    Join multiple results from a selector.
    >>> Join(' - ', '//div/p')  # doctest: +SKIP

    >>> Join(pattern=', ').filter([u"Oui", u"bonjour", ""]) == u"Oui, bonjour"
    True
    >>> Join(pattern='-').filter([u"Au", u"revoir", ""]) == u"Au-revoir"
    True
    >>> Join(pattern='-').filter([]) == u""
    True
    >>> Join(pattern='-', default=u'empty').filter([]) == u'empty'
    True
    """
    def __init__(self, pattern, selector=None, textCleaner=CleanText, newline=False, addBefore='', addAfter='', default=_NO_DEFAULT):
        super(Join, self).__init__(selector, default=default)
        self.pattern = pattern
        self.textCleaner = textCleaner
        self.newline = newline
        self.addBefore = addBefore
        self.addAfter = addAfter

    @debug()
    def filter(self, el):
        items = [self.textCleaner.clean(e) for e in el]
        items = [item for item in items if item]

        if self.newline:
            items = ['%s\r\n' % item for item in items]

        result = self.pattern.join(items)

        if self.addBefore:
            result = '%s%s' % (self.addBefore, result)

        if self.addAfter:
            result = '%s%s' % (result, self.addAfter)

        if not result and self.default is not _NO_DEFAULT:
            return self.default
        return result


class MultiJoin(MultiFilter):
    """
    Join multiple filters.
    >>> MultiJoin(Field('field1'), Field('field2'))  # doctest: +SKIP

    >>> MultiJoin(pattern=u', ').filter([u"Oui", u"bonjour", ""]) == u"Oui, bonjour"
    True
    >>> MultiJoin(pattern=u'-').filter([u"Au", u"revoir", ""]) == u"Au-revoir"
    True
    >>> MultiJoin(pattern=u'-').filter([]) == u""
    True
    >>> MultiJoin(pattern=u'-', default=u'empty').filter([]) == u'empty'
    True
    >>> MultiJoin(pattern=u'-').filter([1, 2, 3]) == u'1-2-3'
    True
    """
    def __init__(self, *args, **kwargs):
        self.pattern = kwargs.pop('pattern', ', ')
        super(MultiJoin, self).__init__(*args, **kwargs)

    @debug()
    def filter(self, values):
        values = [str(v) for v in values if v]
        if not values and self.default is not _NO_DEFAULT:
            return self.default
        return self.pattern.join(values)


class Eval(MultiFilter):
    """
    Evaluate a function with given 'deferred' arguments.

    >>> F = Field; Eval(lambda a, b, c: a * b + c, F('foo'), F('bar'), F('baz')) # doctest: +SKIP
    >>> Eval(lambda x, y: x * y + 1).filter([3, 7])
    22

    Example::

        obj_ratio = Eval(lambda x: x / 100, Env('percentage'))
    """

    def __init__(self, func, *args):
        """
        :param func: function to apply to all filters. The function should
                     accept as many args as there are filters passed to
                     Eval.
        """
        super(Eval, self).__init__(*args)
        self.func = func

    @debug()
    def filter(self, values):
        return self.func(*values)


class QueryValue(Filter):
    """
    Extract the value of a parameter from an URL with a query string.

    >>> from lxml.html import etree
    >>> from .html import Link
    >>> f = QueryValue(Link('//a'), 'id')
    >>> f(etree.fromstring('<html><body><a href="https://example.org/view?id=1234"></a></body></html>')) == u'1234'
    True
    """
    def __init__(self, selector, key, default=_NO_DEFAULT):
        super(QueryValue, self).__init__(selector, default=default)
        self.querykey = key

    @debug()
    def filter(self, url):
        qs = parse_qs(urlparse(url).query)
        if not qs.get(self.querykey):
            return self.default_or_raise(ItemNotFound('Key %s not found' % self.querykey))
        if len(qs[self.querykey]) > 1:
            raise FilterError('More than one value for key %s' % self.querykey)
        return qs[self.querykey][0]


class Coalesce(MultiFilter):
    """
    Returns the first value that is not falsy,
    or default if all values are falsy.
    """
    @debug()
    def filter(self, values):
        for value in values:
            # Accept '0.00' as valid value for numeric elements
            if value or (isinstance(value, Number) and not empty(value)):
                return value
        return self.default_or_raise(FilterError('All falsy and no default.'))



# Copyright(C) 2013
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import BaseObject, Capability, StringField

__all__ = ['SongLyrics', 'CapLyrics']


class SongLyrics(BaseObject):
    """
    Song lyrics object.
    """
    title = StringField('Title of the song')
    artist = StringField('Artist of the song')
    content = StringField('Lyrics of the song')


class CapLyrics(Capability):
    """
    Lyrics websites.
    """

    def iter_lyrics(self, criteria, pattern):
        """
        Search lyrics by artist or by song
        and iterate on results.

        :param criteria: 'artist' or 'song'
        :type criteria: str
        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`SongLyrics`]
        """
        raise NotImplementedError()

    def get_lyrics(self, _id):
        """
        Get a lyrics object from an ID.

        :param _id: ID of lyrics
        :type _id: str
        :rtype: :class:`SongLyrics`
        """
        raise NotImplementedError()



# Copyright(C) 2015 Christophe Lampin
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal

from woob.capabilities.base import empty
from woob.capabilities.shop import CapShop, Order, Item
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter

__all__ = ['AppShop']


class OrdersFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'total')

    def start_format(self, **kwargs):
        self.output('               Id                Date         Total   ')
        self.output('-----------------------------+------------+-----------')

    def format_obj(self, obj, alias):
        date = obj.date.strftime('%Y-%m-%d') if not empty(obj.date) else ''
        total = obj.total or Decimal('0')
        result = '%s  %s  %s' % (self.colored('%-28s' % obj.fullid, 'yellow'),
                                  self.colored('%-10s' % date, 'blue'),
                                  self.colored('%9.2f' % total, 'green'))

        return result

    def flush(self):
        self.output('----------------------------+------------+-----------')


class ItemsFormatter(IFormatter):
    MANDATORY_FIELDS = ('label', 'url', 'price')

    def start_format(self, **kwargs):
        self.output('                                    Label                                                           Url                       Price   ')
        self.output('---------------------------------------------------------------------------+---------------------------------------------+----------')

    def format_obj(self, obj, alias):
        price = obj.price or Decimal('0')
        result = '%s  %s  %s' % (self.colored('%-75s' % obj.label[:75], 'yellow'),
                                  self.colored('%-43s' % obj.url, 'magenta'),
                                  self.colored('%9.2f' % price, 'green'))

        return result

    def flush(self):
        self.output('---------------------------------------------------------------------------+---------------------------------------------+----------')


class PaymentsFormatter(IFormatter):
    MANDATORY_FIELDS = ('date', 'method', 'amount')

    def start_format(self, **kwargs):
        self.output('   Date          Method        Amount  ')
        self.output('-----------+-----------------+----------')

    def format_obj(self, obj, alias):
        date = obj.date.strftime('%Y-%m-%d') if not empty(obj.date) else ''
        amount = obj.amount or Decimal('0')
        result = '%s  %s  %s' % (self.colored('%-10s' % date, 'blue'),
                                  self.colored('%-17s' % obj.method, 'yellow'),
                                  self.colored('%9.2f' % amount, 'green'))

        return result

    def flush(self):
        self.output('-----------+-----------------+----------')


class AppShop(ReplApplication):
    APPNAME = 'shop'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2015 Christophe Lampin'
    DESCRIPTION = 'Console application to obtain details and status of e-commerce orders.'
    SHORT_DESCRIPTION = "obtain details and status of e-commerce orders"
    CAPS = CapShop
    COLLECTION_OBJECTS = (Order, )
    EXTRA_FORMATTERS = {'orders':   OrdersFormatter,
                        'items':   ItemsFormatter,
                        'payments':   PaymentsFormatter
                        }
    DEFAULT_FORMATTER = 'table'
    COMMANDS_FORMATTERS = {'orders':    'orders',
                           'items':     'items',
                           'payments':  'payments',
                           'ls':        'orders',
                           }

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    @defaultcount(10)
    def do_orders(self, line):
        """
        orders [BACKEND_NAME]

        Get orders of a backend.
        If no BACKEND_NAME given, display all orders of all backends.
        """
        if len(line) > 0:
            backend_name = line
        else:
            backend_name = None

        self.do_count(str(self.options.count))  # Avoid raise of MoreResultsAvailable
        l = []
        for order in self.do('iter_orders', backends=backend_name):
            l.append(order)

        self.start_format()
        for order in sorted(l, self.comp_object):
            self.format(order)

    # Order by date DESC
    def comp_object(self, obj1, obj2):
        if obj1.date == obj2.date:
            return 0
        elif obj1.date < obj2.date:
            return 1
        else:
            return -1

    def do_items(self, id):
        """
        items [ID]

        Get items of orders.
        """
        l = []
        id, backend_name = self.parse_id(id, unique_backend=True)

        if not id:
            print('Error: please give a order ID (hint: use orders command)', file=self.stderr)
            return 2
        else:
            l.append((id, backend_name))

        for id, backend in l:
            names = (backend,) if backend is not None else None
            # TODO: Use specific formatter
            mysum = Item()
            mysum.label = "Sum"
            mysum.url = "Generated by shop"
            mysum.price = Decimal("0.")

            self.start_format()
            for item in self.do('iter_items', id, backends=names):
                self.format(item)
                mysum.price = item.price + mysum.price

            self.format(mysum)

    def do_payments(self, id):
        """
        payments [ID]

        Get payments of orders.
        If no ID given, display payment of all backends.
        """

        id, backend_name = self.parse_id(id, unique_backend=True)

        if not id:
            print('Error: please give a order ID (hint: use orders command)', file=self.stderr)
            return 2

        self.start_format()
        for payment in self.do('iter_payments', id, backends=backend_name):
            self.format(payment)
